import numpy
from optimeed.core.tools import printIfShown, SHOW_WARNING
from optimeed.core.additional_tools import convert_color_with_alpha


class Data:
    """This class is used  to store informations necessary to plot a 2D graph. It has to be combined with a gui to be useful (ex. pyqtgraph)"""
    def __init__(self, x: list, y: list, x_label='', y_label='',
                 legend='', is_scattered=False,
                 transfo_x=lambda selfData, x: x, transfo_y=lambda selfData, y: y,
                 xlim=None, ylim=None,
                 permutations=None, sort_output=False,
                 color=None, alpha=255,
                 symbol='o', symbolsize=8, fillsymbol=True, outlinesymbol=1.8,
                 linestyle='-', width=2, meta=None):

        """
        :param x_label: label of x axis
        :param x: x (list) coordinates. If None => it will be the indices
        :param y_label: label of y axis
        :param y: y (list) coordinates.
        :param color: Color to use for the trace (either rgb tuple or matlab-like diminutif (e.g.:'r')
        :param alpha: set transparency (0 to 255)
        :param legend: legend associated with the data
        :param linestyle: style of the line linking points of the trace
        :param is_scattered: boolean. If True: plot will be scattered
        :param sort_output: boolean. If True: x axis will be sorted to ascending number
        :param symbol: symbol to display the points of the trace (e.g.: 'o', 't1', ...)
        :param fillsymbol: paint a solid symbol or not (bool)
        :param outlinesymbol: make the outline of the symbol darker (>1) or clearer (<1). No outline = 1
        :param transfo_x: Method applied to all the x coordinates. Useful to change units for instance
        :param transfo_y: Method applied to all the y coordinates. Useful to change units for instance
        :param xlim: [x_min, x_max] view box
        :param ylim: [y_min, y_max] view box
        :param meta: list (or None if None) of the same length than y. Used for tikz meta.
        """
        if x is None:
            x = []
        if y is None:
            y = []
        self.x = x
        self.x_label = x_label
        self.y = y
        self.y_label = y_label
        self.legend = legend
        self.is_scattered = is_scattered
        self.sort_output = sort_output
        self.symbol = symbol
        self.fillsymbol = fillsymbol
        self.outlinesymbol = outlinesymbol
        self.color = color
        self.alpha = alpha
        self.transfo_x = transfo_x
        self.transfo_y = transfo_y
        self.xlim = xlim
        self.ylim = ylim
        self.points_to_plot = None  # None Value => plot all, [] => plot None
        self.linestyle = linestyle
        self.permutations = None
        self.set_permutations(permutations)
        self.width = width
        self.symbolsize = symbolsize
        self.meta = meta

    def set_kwargs(self, kwargs):
        """Set a kwarg after creation of the class"""
        for kwarg in kwargs:
            setattr(self, kwarg, kwargs[kwarg])

    def set_data(self, x: list, y: list):
        """Overwrites current datapoints with new set"""
        self.x = x
        self.y = y

    def set_meta(self, meta):
        """Set associated 'Z' data"""
        self.meta = meta

    def get_x(self):
        """Get x coordinates of datapoints"""
        the_list = self.x
        if len(self.y) != len(self.x):
            the_list = range(len(self.y))
        return the_list

    def get_symbolsize(self):
        """Get size of the symbols"""
        return self.symbolsize

    def symbol_isfilled(self):
        """Check if symbols has to be filled or not"""
        return self.fillsymbol

    def get_symbolOutline(self):
        """Get color factor of outline of symbols"""
        return self.outlinesymbol

    def get_length_data(self):
        """Get number of points"""
        return max(len(self.x), len(self.y))

    def get_xlim(self):
        """Get x limits of viewbox"""
        return self.xlim

    def get_ylim(self):
        """Get y limits of viewbox"""
        return self.ylim

    def get_y(self):
        """Get y coordinates of datapoints"""
        return self.y

    def get_meta(self):
        """Get associated 'Z' data"""
        if self.meta is None:
            return None
        if len(self.get_y()) == len(self.meta):
            return self.meta
        return None

    def get_color(self):
        """Get color of the line, without transformation"""
        return self.color

    def get_color_alpha(self):
        """Get color of the line. Return r, g, b in 0, 255 scale"""
        return convert_color_with_alpha(self.color, self.alpha)

    def get_alpha(self):
        """Get opacity"""
        return self.alpha

    def get_width(self):
        """Get width of the line"""
        return self.width

    def get_number_of_points(self):
        """Get number of points"""
        return len(self.get_permutations())

    def get_plot_data(self):
        """
        Call this method to get the x and y coordinates of the points that have to be displayed.
        => After transformation, and after permutations.

        :return: x (list), y (list)
        """
        x = self.get_x()
        y = self.get_y()

        x = [self.transfo_x(self, i) for i in x]
        y = [self.transfo_y(self, i) for i in y]
        permutations = self.get_permutations(x=x)
        x, y = [x[perm_i] for perm_i in permutations], [y[perm_i] for perm_i in permutations]
        min_length = min(len(x), len(y))
        return x[:min_length], y[:min_length]

    def get_plot_meta(self, x, y):
        """
        Call this method to get the z coordinates of the points that been displayed.
        => After transformation, and after permutations.

        :return: z (list)
        """
        z = self.get_meta()
        if z is None:
            return z

        permutations = self.get_permutations(x=x)
        min_length = min(len(x), len(y))
        z = [z[perm_i] for perm_i in permutations]
        return z[:min_length]

    def get_permutations(self, x=None):
        """Return the transformation 'permutation':
        xplot[i] = xdata[permutation[i]]
        """
        indices_to_plot = self.get_indices_points_to_plot()

        if x is None:
            x = self.get_x()

        if self.permutations is not None:
            permutations = self.permutations
        elif self.sort_output:
            permutations = list(numpy.argsort(x))
        else:
            permutations = range(len(x))
            if len(indices_to_plot) == self.get_length_data():
                return permutations
        return [permutations[index] for index in indices_to_plot]

    def get_invert_permutations(self):
        """Return the inverse of permutations:
        xdata[i] = xplot[revert[i]]
        """
        return numpy.argsort(self.get_permutations())

    def get_dataIndex_from_graphIndex(self, index_graph_point):
        """
        From an index given in graph, recovers the index of the data.

        :param index_graph_point: Index in the graph
        :return: index of the data
        """
        return self.get_permutations()[index_graph_point]

    def get_dataIndices_from_graphIndices(self, index_graph_point_list):
        """
        Same as get_dataIndex_from_graphIndex but with a list in entry.
        Can (?) improve performances for huge dataset.

        :param index_graph_point_list: List of Index in the graph
        :return: List of index of the data
        """
        permutations = self.get_permutations()
        return [permutations[index] for index in index_graph_point_list]

    def get_graphIndex_from_dataIndex(self, index_data):
        """
        From an index given in the data, recovers the index of the graph.

        :param index_data: Index in the data
        :return: index of the graph
        """
        return self.get_permutations().index(index_data)

    def get_graphIndices_from_dataIndices(self, index_data_list):
        """
        Same as get_graphIndex_from_dataIndex but with a list in entry.
        Can (?) improve performances for huge dataset.

        :param index_data_list: List of Index in the data
        :return: List of index of the graph
        """
        invert_permutation = self.get_invert_permutations()
        return [invert_permutation[index] for index in index_data_list]

    def set_permutations(self, permutations):
        """
        Set permutations between datapoints of the trace

        :param permutations: list of indices to plot (example: [0, 2, 1] means that the first point will be plotted, then the third, then the second one)
        """
        if permutations is not None:
            if len(self.get_x()) == len(permutations):
                if self.sort_output:
                    print("Warning : Permutations due to flag 'sort_output' are overridden by user")
                self.permutations = permutations
            else:
                print("Error : Permutations have not the same length as the data")

    def get_x_label(self):
        """ Get x label of the trace """
        return self.x_label

    def get_y_label(self):
        """ Get y label of the trace """
        return self.y_label

    def get_legend(self):
        """ Get name of the trace """
        return self.legend

    def get_symbol(self):
        """ Get symbol """
        return self.symbol

    def add_point(self, x, y):
        """Add point(s) to trace (inputs can be list or numeral)"""
        if not isinstance(x, list):
            x = [x]
        if not isinstance(y, list):
            y = [y]
        self.x.extend(x)
        self.y.extend(y)

    def delete_point(self, index_point):
        """Delete a point from the datapoints"""
        if len(self.x):
            del self.x[index_point]
            del self.y[index_point]
        else:
            del self.y[index_point]

    def isScattered(self):
        """Check if plot is scatteded"""
        return self.is_scattered or len(self.get_y()) == 1

    def set_indices_points_to_plot(self, indices):
        """Set indices points to plot"""
        self.points_to_plot = indices

    def get_indices_points_to_plot(self):
        """Get indices points to plot"""
        indices = self.points_to_plot
        if indices is None:
            indices = range(self.get_length_data())
        return indices

    def get_linestyle(self):
        """Get linestyle"""
        return str(self.linestyle)

    def __str__(self):
        theStr = self.x_label + '\t' + self.y_label + '\n'
        theX = self.get_x()
        theY = self.get_y()
        for i in range(self.get_length_data()):
            theStr += str(theX[i]) + '\t\t\t' + str(theY[i]) + '\n'
        return theStr

    def export_str(self):
        """Method to save the points constituting the trace"""
        theStr = "# $NEW TRACE 2D$\n"
        theStr += "# $LEGEND$: ${}$\n".format(self.get_legend())
        theStr += "# $LABEL X$: ${}$\n".format(self.get_x_label())
        theStr += "# $LABEL Y$: ${}$\n".format(self.get_y_label())
        theStr += "# $BEGIN DATA$\n"
        x, y = self.get_plot_data()
        for i in range(len(x)):
            theStr += "{}\t{}\n".format(x[i], y[i])
        theStr += "# $END DATA$\n"
        return theStr

    def set_color(self, theColor):
        """Set trace color"""
        self.color = theColor

    def set_legend(self, theLegend):
        """Set legend"""
        self.legend = theLegend


class Graph:
    """Simple graph container that contains several traces"""
    def __init__(self):
        self.traces = dict()
        self.curr_id = 0

    def add_trace(self, data):
        """Add a trace to the graph

        :param data: :class:`~Data`
        :return: id of the created trace
        """
        idTrace = self.curr_id
        self.traces[idTrace] = data
        self.curr_id += 1
        return idTrace

    def remove_trace(self, idTrace):
        """Delete a trace from the graph

        :param idTrace: id of the trace to delete
        """
        try:
            del self.traces[idTrace]
        except KeyError:
            printIfShown("Key {} not found".format(idTrace), SHOW_WARNING)

    def get_trace(self, idTrace) -> Data:
        """Get data object of idTrace

        :param idTrace: id of the trace to get
        :return: :class:`~Data`
        """
        return self.traces[idTrace]

    def get_all_traces(self):
        """Get all the traces id of the graph"""
        return self.traces

    def get_all_traces_ids(self):
        """Get all the traces id of the graph
        :return: list of id graphs
        """
        return list(self.traces.keys())

    def export_str(self):
        theStr = "# $NEW GRAPH$\n\n"
        for idTrace in self.get_all_traces():
            theStr += self.get_trace(idTrace).export_str()
            theStr += "\n"
        return theStr


class Graphs:
    """Contains several :class:`Graph`"""
    def __init__(self):
        self.graphs = dict()
        self.curr_id = 0
        self.updateMethods = set()

    def updateChildren(self):
        for updateMethod in self.updateMethods:
            updateMethod()

    def add_trace_firstGraph(self, data, updateChildren=True):
        """
        Same as add_trace, but only if graphs has only one id
        :param data:
        :param updateChildren:
        :return:
        """
        all_ids = self.get_all_graphs_ids()
        if len(all_ids) == 1:
            return self.add_trace(all_ids[0], data, updateChildren=updateChildren)
        printIfShown("Cannot add trace .. graphs are multiple", SHOW_WARNING)
        return None

    def add_trace(self, idGraph, data, updateChildren=True):
        """Add a trace to the graph

        :param idGraph: id of the graph
        :param data: :class:`~Data`
        :param updateChildren: Automatically calls callback functions
        :return: id of the created trace
        """
        idTrace = self.get_graph(idGraph).add_trace(data)
        if updateChildren:
            self.updateChildren()
        return idTrace

    def remove_trace(self, idGraph, idTrace, updateChildren=True):
        """Remove the trace from the graph

        :param idGraph: id of the graph
        :param idTrace: id of the trace to remove
        :param updateChildren: Automatically calls callback functions
        """
        self.get_graph(idGraph).remove_trace(idTrace)
        if updateChildren:
            self.updateChildren()

    def get_first_graph(self):
        """Get id of the first graph

        :return: id of the first graph
        """
        return self.get_graph(self.get_all_graphs_ids()[0])

    def get_graph(self, idGraph):
        """Get graph object at idgraph

        :param idGraph: id of the graph to get
        :return: :class:`~Graph`
        """
        return self.graphs[idGraph]

    def get_all_graphs_ids(self):
        """Get all ids of the graphs

        :return: list of id graphs
        """
        return list(self.graphs.keys())

    def get_all_graphs(self):
        """Get all graphs. Return dict {id: :class:`~Graph`}"""
        return self.graphs

    def add_graph(self, updateChildren=True):
        """Add a new graph

        :return: id of the created graph
        """
        idGraph = self.curr_id
        self.graphs[idGraph] = Graph()
        self.curr_id += 1
        if updateChildren:
            self.updateChildren()
        return idGraph

    def remove_graph(self, idGraph):
        """Delete a graph

        :param idGraph: id of the graph to delete
        """
        try:
            del self.graphs[idGraph]
        except KeyError:
            printIfShown("Key {} not found".format(idGraph), SHOW_WARNING)
        self.updateChildren()

    def add_update_method(self, childObject):
        """Add a callback each time a graph is modified.

        :param childObject: method without arguments
        """
        self.updateMethods.add(childObject)

    def export_str(self):
        """Export all the graphs in text

        :return: str"""
        theStr = ""
        for graphId in self.get_all_graphs_ids():
            theStr += self.get_graph(graphId).export_str()
            theStr += "\n\n\n"
        return theStr

    def merge(self, otherGraphs):
        # curr_id = 0
        # mappings = [{}]*2
        # all_graphs_1 = self.get_all_graphs()
        # all_graphs_2 = otherGraphs.get_all_graphs()
        #
        # new_graphs = dict()
        #
        # for index, all_graphs in enumerate([all_graphs_1, all_graphs_2]):
        #     for graphId in all_graphs_1:
        #         new_graphs[curr_id] = all_graphs[graphId]
        #         mappings[index][graphId] = curr_id
        #         curr_id += 1
        # self.graphs = new_graphs
        id_1 = self.get_all_graphs_ids()
        id_2 = otherGraphs.get_all_graphs_ids()

        not_in_1 = list(set(id_2) - set(id_1))
        in_1 = list(set(id_1) & set(id_2))

        # Add traces belonging to second graphs to first
        for id_graph in in_1:
            for trace_data in otherGraphs.get_graph(id_graph).get_all_traces().values():
                self.add_trace(id_graph, trace_data, updateChildren=False)
        # Create new graphs and add traces belonging to second graphs to first
        for id_graph in not_in_1:
            new_id = self.add_graph(updateChildren=False)
            for trace_data in otherGraphs.get_graph(id_graph).get_all_traces().values():
                self.add_trace(new_id, trace_data, updateChildren=False)
        self.updateChildren()

    def reset(self):
        self.graphs = dict()
        self.updateChildren()

    def is_empty(self):
        return len(self.get_all_graphs()) == 0



from optimeed.core.collection import AutosaveStruct, ListDataStruct, Performance_ListDataStruct, SingleObjectSaveLoad
from optimeed.core.tools import create_unique_dirname, getPath_workspace, indentParagraph
from optimeed.core.ansi2html import Ansi2HTMLConverter
from typing import List
import numpy as np
import os


class OptiHistoric:
    """Contains all the points that have been evaluated"""

    class _pointData:
        time: float
        objectives: List[float]
        constraints: List[float]

        def __init__(self, currTime, objectives, constraints):
            self.time = currTime
            self.objectives = objectives
            self.constraints = constraints

    class _LogParams:  # Keep track of evaluated parameters
        def __init__(self):
            self.params = None
            self.curr_row = 0

        def add_parameters(self, params):
            if self.params is None:
                self.params = np.zeros((int(1e6), len(params)))

            try:
                self.params[self.curr_row, :] = params[:]
            except IndexError:
                self.params = np.vstack((self.params, np.zeros((int(1e6), len(params)))))
                self.params[self.curr_row, :] = params[:]

            self.curr_row += 1

        def get_rows_indices(self, list_of_params):
            truncated_matrix = self.params[:self.curr_row+1, :]
            list_of_indices = []
            for params in list_of_params:
                try:
                    index = np.where(np.all(truncated_matrix == np.array(params), axis=1))[0][0]
                    list_of_indices.append(index)
                except IndexError:
                    return False, list_of_indices
            return True, list_of_indices

    def __init__(self, optiname="opti", autosave_timer=60*5, autosave=True, create_new_directory=True, performance_datastruct=True):
        super().__init__()

        foldername = "{}/{}".format(getPath_workspace(), optiname)
        if create_new_directory:
            foldername = create_unique_dirname(foldername)
        self.foldername = foldername

        if performance_datastruct:
            self.log_devices = AutosaveStruct(Performance_ListDataStruct(), filename=os.path.join(self.foldername, "autosaved"))
        else:
            self.log_devices = AutosaveStruct(ListDataStruct(compress_save=False), filename=os.path.join(self.foldername, "autosaved"))
        self.log_opti = AutosaveStruct(ListDataStruct(compress_save=True), filename=os.path.join(self.foldername, "logopti"))
        self.log_convergence = AutosaveStruct(ListDataStruct(compress_save=False), filename=os.path.join(self.foldername, "optiConvergence"))
        self.log_params = self._LogParams()

        if autosave:
            self.log_devices.start_autosave(autosave_timer)
            self.log_opti.start_autosave(autosave_timer)
            self.log_convergence.start_autosave(autosave_timer)

        self.autosave = autosave
        self.results = AutosaveStruct(ListDataStruct(compress_save=True), filename=os.path.join(self.foldername, "results"))

    def log_after_evaluation(self, returned_values: dict):
        """Save the output of evaluate to optiHistoric. This function should be called by the optimizer IN a process safe context."""
        self.log_params.add_parameters(returned_values["params"])
        self.log_devices.get_datastruct().add_data(returned_values["device"])
        self.log_opti.get_datastruct().add_data(self._pointData(returned_values["time"], returned_values["objectives"], returned_values["constraints"]))

    def set_results(self, devicesList):
        self.results.get_datastruct().set_data(devicesList)

    def get_best_devices_without_reevaluating(self, list_of_best_params):
        success, indices_results = self.log_params.get_rows_indices(list_of_best_params)
        solutions = [self.log_devices.get_datastruct().get_data_at_index(index) for index in indices_results]
        return success, solutions

    def set_convergence(self, theConvergence):
        self.log_convergence.get_datastruct().set_data([theConvergence])

    def save(self):
        if self.autosave:
            for struct in [self.log_devices, self.log_opti, self.log_convergence, self.results]:
                struct.stop_autosave()
                struct.save()

    def get_convergence(self):
        """ :return: convergence :class:`~optimeed.optimize.optiAlgorithms.convergence.interfaceConvergence.InterfaceConvergence` """
        return self.log_convergence.get_datastruct()

    def get_devices(self):
        """ :return: List of devices (ordered by evaluation number) """
        return self.log_devices.get_datastruct()

    def get_logopti(self):
        """:return: Log optimization (to check the convergence) """
        return self.log_opti.get_datastruct()

    def start(self, optimization_parameters):
        """Function called upon starting the optimization. Create folders."""

        if self.autosave:
            SingleObjectSaveLoad.save(optimization_parameters, os.path.join(self.foldername, "optimization_parameters.json"))

            # Create summary file
            optimization_info = ''
            optimization_info += '--------------------------------------------------------------------\n'
            optimization_info += 'OBJECTIVES:\n'
            for objective in optimization_parameters.get_objectives():
                optimization_info += indentParagraph("• {} | {}".format(objective, type(objective)), indent_level=1)

            optimization_info += 'CONSTRAINTS:\n'
            for constraint in optimization_parameters.get_constraints():
                optimization_info += indentParagraph("• {} | {}".format(constraint, type(constraint)), indent_level=1)

            optimization_info += 'OPTIMIZATION VARIABLES :\n'
            for optiVariable in optimization_parameters.get_optivariables():
                optimization_info += indentParagraph('• ' + str(optiVariable), indent_level=1)

            optimization_info += 'OPTIMIZATION ALGORITHM:\n'
            optimization_info += indentParagraph("{} | {}".format(optimization_parameters.get_optialgorithm(), type(optimization_parameters.get_optialgorithm())), indent_level=1)

            optimization_info += 'CHARACTERIZATION SCHEME:\n'
            optimization_info += indentParagraph("{} | {}".format(optimization_parameters.get_charac(), type(optimization_parameters.get_charac())), indent_level=1)

            optimization_info += 'M2P:\n'
            optimization_info += indentParagraph("{} | {}".format(optimization_parameters.get_M2P(), type(optimization_parameters.get_M2P())), indent_level=1)

            with open(self.foldername + "/summary.html", 'w', encoding='utf-8') as myFile:
                conv = Ansi2HTMLConverter(dark_bg=True)
                myFile.write(conv.convert(optimization_info))





#!/usr/bin/python
"""
For testing rapid updates of ScatterPlotItem under various conditions.
"""

from time import perf_counter

import numpy as np

import pyqtgraph as pg
import pyqtgraph.parametertree as ptree
from pyqtgraph.Qt import QtCore, QtWidgets

translate = QtCore.QCoreApplication.translate

app = pg.mkQApp()
param = ptree.Parameter.create(name=translate('ScatterPlot', 'Parameters'), type='group', children=[
    dict(name='paused', title=translate('ScatterPlot', 'Paused:    '), type='bool', value=False),
    dict(name='count', title=translate('ScatterPlot', 'Count:    '), type='int', limits=[1, None], value=500, step=100),
    dict(name='size', title=translate('ScatterPlot', 'Size:    '), type='int', limits=[1, None], value=10),
    dict(name='randomize', title=translate('ScatterPlot', 'Randomize:    '), type='bool', value=False),
    dict(name='pxMode', title='pxMode:    ', type='bool', value=True),
    dict(name='useCache', title='useCache:    ', type='bool', value=True),
    dict(name='mode', title=translate('ScatterPlot', 'Mode:    '), type='list', limits={translate('ScatterPlot', 'New Item'): 'newItem', translate('ScatterPlot', 'Reuse Item'): 'reuseItem', translate('ScatterPlot', 'Simulate Pan/Zoom'): 'panZoom', translate('ScatterPlot', 'Simulate Hover'): 'hover'}, value='reuseItem'),
])
for c in param.children():
    c.setDefault(c.value())

pt = ptree.ParameterTree(showHeader=False)
pt.setParameters(param)
p = pg.PlotWidget()
splitter = QtWidgets.QSplitter()
splitter.addWidget(pt)
splitter.addWidget(p)
splitter.show()

data = {}
item = pg.ScatterPlotItem()
hoverBrush = pg.mkBrush('y')
ptr = 0
lastTime = perf_counter()
fps = None
timer = QtCore.QTimer()


def mkDataAndItem():
    global data, fps
    scale = 100
    data = {
        'pos': np.random.normal(size=(50, param['count']), scale=scale),
        'pen': [pg.mkPen(x) for x in np.random.randint(0, 256, (param['count'], 3))],
        'brush': [pg.mkBrush(x) for x in np.random.randint(0, 256, (param['count'], 3))],
        'size': (np.random.random(param['count']) * param['size']).astype(int)
    }
    data['pen'][0] = pg.mkPen('w')
    data['size'][0] = param['size']
    data['brush'][0] = pg.mkBrush('b')
    bound = 5 * scale
    p.setRange(xRange=[-bound, bound], yRange=[-bound, bound])
    mkItem()


def mkItem():
    global item
    item = pg.ScatterPlotItem(pxMode=param['pxMode'], **getData())
    item.opts['useCache'] = param['useCache']
    p.clear()
    p.addItem(item)


def getData():
    pos = data['pos']
    pen = data['pen']
    size = data['size']
    brush = data['brush']
    if not param['randomize']:
        pen = pen[0]
        size = size[0]
        brush = brush[0]
    return dict(x=pos[ptr % 50], y=pos[(ptr + 1) % 50], pen=pen, brush=brush, size=size)


def update():
    global ptr, lastTime, fps
    mode = param['mode']
    if mode == 'newItem':
        mkItem()
    elif mode == 'reuseItem':
        item.setData(**getData())
    elif mode == 'panZoom':
        item.viewTransformChanged()
        item.update()
    elif mode == 'hover':
        pts = item.points()
        old = pts[(ptr - 1) % len(pts)]
        new = pts[ptr % len(pts)]
        item.pointsAt(new.pos())
        old.resetBrush()  # reset old's brush before setting new's to better simulate hovering
        new.setBrush(hoverBrush)

    ptr += 1
    now = perf_counter()
    dt = now - lastTime
    lastTime = now
    if fps is None:
        fps = 1.0 / dt
    else:
        s = np.clip(dt * 3., 0, 1)
        fps = fps * (1 - s) + (1.0 / dt) * s
    p.setTitle('%0.2f fps' % fps)
    p.repaint()
    # app.processEvents()  # force complete redraw for every plot


mkDataAndItem()
for name in ['count', 'size']:
    param.child(name).sigValueChanged.connect(mkDataAndItem)
for name in ['useCache', 'pxMode', 'randomize']:
    param.child(name).sigValueChanged.connect(mkItem)
param.child('paused').sigValueChanged.connect(lambda _, v: timer.stop() if v else timer.start())
timer.timeout.connect(update)
timer.start(0)

if __name__ == '__main__':
    pg.exec()



from .onclickInterface import OnclickInterface


class Onclick_delete(OnclickInterface):
    """On Click: Delete the points from the graph"""

    def __init__(self, theDataLink):
        """

        :param theDataLink: :class:`~optimeed.visualize.high_level.LinkDataGraph.LinkDataGraph`
        """
        super().__init__()
        self.theDataLink = theDataLink

    def graph_clicked(self, _theGraphVisual, index_graph, index_trace, indices_points):
        self.theDataLink.delete_clicked_items(index_graph, index_trace, indices_points)

    def get_name(self):
        return "Delete points"




from optimeed.core.tools import get_ND_pareto
from optimeed.optimize.optiAlgorithms.convergence.hypervolume import HyperVolume
import numpy as np
from .interfaceConvergence import InterfaceConvergence
from optimeed.core import Graphs, Data
from typing import Dict, List
from math import floor
from optimeed.core import printIfShown, SHOW_WARNING
import traceback


class EvolutionaryConvergence(InterfaceConvergence):
    """convergence class for population-based algorithm"""
    objectives_per_step: Dict[int, List[List[float]]]
    constraints_per_step: Dict[int, List[List[float]]]
    paretos_per_step: Dict[int, List[List[float]]]
    hypervolume_per_step: Dict[int, List[float]]

    def __init__(self):
        self.objectives_per_step = dict()
        self.constraints_per_step = dict()
        self.paretos_per_step = dict()
        self.curr_step = 0

    def set_curr_step(self, theObjectives_list, theConstraints_list):
        # Todo: Speed up method
        theStep = self.curr_step

        self.objectives_per_step[theStep] = theObjectives_list
        self.constraints_per_step[theStep] = theConstraints_list
        # Find new pareto
        objectives_respect_constraint = self.get_objectives_respect_constraint(theObjectives_list, theConstraints_list)
        prev_pareto = self.paretos_per_step[theStep-1][:] if theStep > 0 else list()
        try:
            new_pareto, _ = get_ND_pareto(prev_pareto + objectives_respect_constraint)
        except IndexError:
            new_pareto = list()
        self.paretos_per_step[theStep] = new_pareto

        self.curr_step += 1

    @staticmethod
    def get_objectives_respect_constraint(objectives, constraints):
        objectives_respect_constraint = list()
        for i in range(len(objectives)):
            respect_constraints = True
            for constraint in constraints[i]:
                if constraint > 0:
                    respect_constraints = False
                    break
            if respect_constraints:
                objectives_respect_constraint.append(objectives[i])
        return objectives_respect_constraint

    def _extract_N_steps(self, N):
        nbSteps = self.last_step()
        if N is None or nbSteps <= N:
            steps = list(range(nbSteps+1))
        else:
            stepsize = floor(nbSteps / N)
            steps = list(range(0, stepsize * N, stepsize))
            if steps[-1] != nbSteps:
                steps.append(nbSteps)
        return steps

    def get_pareto_convergence(self, max_number_of_points=None):
        if self.paretos_per_step:
            return {step: self.paretos_per_step[step] for step in self._extract_N_steps(max_number_of_points)}
        else:
            printIfShown("Deprecated use, probably from deprecated file.", SHOW_WARNING)
            paretos = dict()
            previous_pareto = list()

            for step in self._extract_N_steps(max_number_of_points):
                try:
                    paretos[step], _ = get_ND_pareto(previous_pareto + self.get_objectives_respect_constraint(self.objectives_per_step[step], self.constraints_per_step[step]))
                except IndexError:
                    paretos[step] = list()
                previous_pareto = paretos[step][:]
            return paretos

    def get_pareto_at_step(self, step):
        if self.paretos_per_step:
            return self.paretos_per_step[step]

        previous_pareto = list()
        pareto = list()
        for step in range(step+1):
            try:
                pareto, _ = get_ND_pareto(previous_pareto + self.get_objectives_respect_constraint(self.objectives_per_step[step], self.constraints_per_step[step]))
            except IndexError:
                pareto = list()
            previous_pareto = pareto[:]
        return pareto

    def get_hypervolume(self, pareto, refPoint=None):
        if not pareto:
            return np.nan, None
        if self.get_nb_objectives() == 1:  # monobj
            return pareto[0][0], None
        if refPoint is None:
            refPoint = self.get_nadir_point(pareto)
        hv = HyperVolume(refPoint)
        return hv.compute(pareto), refPoint

    def get_hypervolume_convergence(self, max_number_of_points):
        paretos = self.get_pareto_convergence(max_number_of_points=max_number_of_points)
        res_dict = dict()
        if len(paretos[self.last_step()]) == 1 and len(paretos[self.last_step()][0]) > 1:  # only one point in last pareto ... => objectives in same direction
            printIfShown("Objectives go in same direction, using sum of objectives instead", SHOW_WARNING)
            for step in self._extract_N_steps(max_number_of_points):
                try:
                    res_dict[step] = np.sqrt(sum([objective**2 for objective in paretos[step][0]]))
                except IndexError:
                    res_dict[step] = 0.0
            return res_dict

        nadir_point = self.get_nadir_point(paretos[self.last_step()])
        for step in self._extract_N_steps(max_number_of_points):
            res_dict[step], _ = self.get_hypervolume(paretos[step], refPoint=nadir_point)
        return res_dict

    def get_nadir_point(self, pareto):
        nb_objectives = self.get_nb_objectives()
        nadir = [-np.inf] * nb_objectives
        if len(pareto) <= 1:  # Only one point in pareto front
            return nadir

        for i in range(nb_objectives):
            try:
                nadir[i] = max([max(nadir[i], objectives[i]) for objectives in pareto])
            except ValueError:
                pass
        return nadir

    def last_step(self):
        if self.objectives_per_step:
            return max(self.objectives_per_step.keys())
        return 0

    def get_nb_objectives(self):
        try:
            return len(self.objectives_per_step[self.last_step()][0])
        except (ValueError, KeyError):
            return 0

    def get_scalar_convergence_evolution(self, max_number_of_points):
        # try:
        hypervolumes = self.get_hypervolume_convergence(max_number_of_points)
        x = list(hypervolumes.keys())
        y = list(hypervolumes.values())
        return x, y
        # except KeyError:
        #     return [], []

    def get_graphs(self, max_number_of_points=None):
        theGraphs = Graphs()
        # noinspection PyBroadException
        try:
            x, y = self.get_scalar_convergence_evolution(max_number_of_points)

            if self.get_nb_objectives() == 1:
                g2 = theGraphs.add_graph()

                theDataPareto = Data(x, y, x_label="Step", y_label="Objective", symbol=None)
                theGraphs.add_trace(g2, theDataPareto)
            else:
                g1 = theGraphs.add_graph()
                theGraphs.add_trace(g1, Data(x, y, x_label="Step", y_label="Hypervolume"))
                n = self.get_nb_objectives()
                for i in range(n):
                    for j in range(i+1,n):
                        g = theGraphs.add_graph()
                        paretos = self.get_pareto_convergence(max_number_of_points=10)
                        for step in paretos:
                            theDataPareto = Data([objectives[i] for objectives in paretos[step]], [objectives[j] for objectives in paretos[step]], sort_output=True, x_label=f"Objective_{i}", y_label=f"Objective_{j}")
                            theGraphs.add_trace(g, theDataPareto)
        except Exception:
            printIfShown("An error in convergence graphs. Error :" + traceback.format_exc(), SHOW_WARNING)

        return theGraphs



from optimeed.optimize.optiAlgorithms import MultiObjective_GA as OptimizationAlgorithm
from optimeed.optimize import run_optimization, Real_OptimizationVariable, InterfaceObjCons, OptimizerSettings, OptiHistoric
import numpy as np


class _Device:
    def __init__(self, fitFunction, nbArgs):
        self.functionArgs = [0.0]*nbArgs
        self.fitFunction = fitFunction


class _Objective(InterfaceObjCons):
    def __init__(self, x_data, y_data, fitCriterion):
        self.x_data = x_data
        self.y_data = y_data
        self.fitCriterion = fitCriterion

    def compute(self, theDevice):
        return self.fitCriterion(theDevice.fitFunction, theDevice.functionArgs, self.x_data, self.y_data)


def leastSquare(function, functionArgs, x_data, y_data):
    """
    Least square calculation (sum (y-ŷ)^2)

    :param function: Function to fit
    :param functionArgs: Arguments of the function
    :param x_data: x-axis coordinates of data to fit
    :param y_data: y-axis coordinates of data to fit
    :return: least squares
    """
    return np.sum((y_data - function(x_data, *functionArgs)) ** 2)


def r_squared(function, functionArgs, x_data, y_data):
    """
    R squared calculation

    :param function: Function to fit
    :param functionArgs: Arguments of the function
    :param x_data: x-axis coordinates of data to fit
    :param y_data: y-axis coordinates of data to fit
    :return: R squared
    """
    return -(1 - (np.sum((y_data - function(x_data, *functionArgs)) ** 2))/(np.sum((y_data - np.mean(y_data)) ** 2)))


def do_fit(fitFunction, x_data, y_data, *args, fitCriterion=leastSquare):
    """
    Main method to fit a function

    :param fitFunction: the function to fit (link to it)
    :param x_data: x-axis coordinates of data to fit
    :param y_data: y-axis coordinates of data to fit
    :param args: for each parameter: [min, max] admissible value
    :param fitCriterion: fit criterion to minimize. Default: least square
    :return: [arg_i_optimal, ...], y estimated, error.
    """
    print(args)
    theDevice = _Device(fitFunction, len(args))
    theAlgo = OptimizationAlgorithm()

    optimizationVariables = list()
    for i, minmax in enumerate(args):
        optimizationVariables.append(Real_OptimizationVariable('functionArgs[{}]'.format(i), *minmax))  #

    listOfObjectives = [_Objective(x_data, y_data, fitCriterion)]
    listOfConstraints = []

    theOptiParameters = OptimizerSettings(theDevice, listOfObjectives, listOfConstraints, optimizationVariables, theOptimizationAlgorithm=theAlgo)
    theOptiHistoric = OptiHistoric(optiname="fit", autosave=False, create_new_directory=False, performance_datastruct=False)

    resultsOpti, convergence = run_optimization(theOptiParameters, theOptiHistoric, max_opti_time_sec=3)

    y_est = fitFunction(x_data, *resultsOpti[0].functionArgs)

    return resultsOpti[0].functionArgs, y_est, y_data-y_est



from PyQt5 import QtWidgets, QtCore
from optimeed.core.collection import ListDataStruct
import os


class CollectionExporterGUI(QtWidgets.QMainWindow):
    """Simple gui that allows to export data"""
    signal_has_exported = QtCore.pyqtSignal()
    signal_has_reset = QtCore.pyqtSignal()

    def __init__(self):
        self.theCollection = ListDataStruct()

        # Create PyQt5 window
        super().__init__()
        self.mainbox = QtWidgets.QWidget()
        self.setCentralWidget(self.mainbox)
        self.mainbox.setLayout(QtWidgets.QVBoxLayout())
        self.mainbox.layout().setContentsMargins(0, 0, 0, 0)

        self.exportButton = QtWidgets.QPushButton('Export Collection', self)
        self.mainbox.layout().addWidget(self.exportButton)
        self.exportButton.clicked.connect(self.exportCollection)

        resetButton = QtWidgets.QPushButton('Reset', self)
        self.mainbox.layout().addWidget(resetButton)
        resetButton.clicked.connect(self.reset)

    def exportCollection(self):
        """Export the collection"""
        dlg = QtWidgets.QFileDialog.getSaveFileName()[0]
        if dlg:
            root, ext = os.path.splitext(dlg)
            filename_collection = root + self.theCollection.get_extension()
            self.theCollection.save(filename_collection)
            self.signal_has_exported.emit()

    def reset(self):
        self.theCollection.reset_data()
        self.signal_has_reset.emit()

    def add_data_to_collection(self, data):
        """
        Add data to the collection to export

        :param data: Whichever type you like
        """
        self.theCollection.add_data(data)
        self.show()

    def set_collection(self, theCollection):
        self.theCollection = theCollection



from abc import ABCMeta, abstractmethod


class InterfaceCharacterization(metaclass=ABCMeta):
    """Interface for the evaluation of a device"""

    @abstractmethod
    def compute(self, theDevice):
        """
        Action to perform to characterize (= compute the objective function) of the device.

        :param theDevice: the device to characterize
        """
        pass

    def __str__(self):
        return "Characterization - No name given"



import math
import numpy as np
from OpenGL.GL import *

from . import quaternions


MODE_ZOOM = 0
MODE_ROTATION = 1
MODE_LIGHT = 2
NUMBER_OF_MODES = 3

CLIC_LEFT = 0
CLIC_RIGHT = 1


class SpecialButtonsMapping:
    def __init__(self):
        self.KEY_LEFT = None
        self.KEY_RIGHT = None
        self.KEY_UP = None
        self.KEY_DOWN = None
        self.KEY_TAB = None
        self.MOUSE_LEFT = None
        self.MOUSE_RIGHT = None


class MyText:
    def __init__(self, color, fontSize, theStr, windowPosition):
        self.color = color
        self.fontSize = fontSize
        self.position = windowPosition
        self.theStr = theStr


class ContextHandler:

    def __init__(self):
        self.rotation_accumulation = (1, 0, 0, 0)
        self.width = 200
        self.height = 200

        self.theDeviceDrawer = None
        self.theDeviceToDraw = None
        self.theSpecialButtonsMapping = None

        # Initial viewPoint and light
        self.__displayScale__ = False
        self.scaleLength = 5

        self.tipAngle, self.viewAngle, self.zoomLevel = 10, 10, 40

        self.must_rotate = True
        self.lightPosition = [-16100.0, 15900.0, 0.0, 1.0]

        self.backgroundColor = [0.3, 0.3, 0.3]

        self.centerPosition = [0, 0, 0]

        self.xMouseOnClic = 0
        self.yMouseOnClic = 0

        self.currentMode = MODE_ROTATION
        self.currentClic = CLIC_LEFT

        try:
            self.VP_MarginW
        except AttributeError:
            self.VP_MarginW = 0
            self.VP_MarginH = 0
            self.VP_W = self.width
            self.VP_H = self.height

    def set_specialButtonsMapping(self, theSpecialButtonsMapping):
        self.theSpecialButtonsMapping = theSpecialButtonsMapping

    def set_deviceDrawer(self, theDeviceDrawer):
        self.theDeviceDrawer = theDeviceDrawer

    def set_deviceToDraw(self, theDeviceToDraw):
        if self.theDeviceToDraw is None:
            self.theDeviceToDraw = theDeviceToDraw
            _, _, self.zoomLevel = self.theDeviceDrawer.get_init_camera(self.theDeviceToDraw)
        else:
            self.theDeviceToDraw = theDeviceToDraw

    def resizeWindowAction(self, new_width, new_height):
        ar_origin = self.width / self.height
        ar_new = new_width / new_height

        scale_w = new_width / self.width
        scale_h = new_height / self.height

        if ar_new > ar_origin:
            scale_w = scale_h
        else:
            scale_h = scale_w

        margin_x = (new_width - self.width * scale_w / ar_origin) / 2
        margin_y = (new_height - self.height * scale_h) / 2
        self.VP_H = int(self.height * scale_h)
        self.VP_W = int(self.width * scale_w / ar_origin)
        self.VP_MarginH = int(margin_y)
        self.VP_MarginW = int(margin_x)

    def mouseWheelAction(self, deltaAngle):
        self.zoomLevel += 0.1 * self.zoomLevel * deltaAngle

        # GLUT.glutPostRedisplay()

    def mouseClicAction(self, button, my_x, y):
        self.xMouseOnClic = my_x
        self.yMouseOnClic = y
        self.currentClic = button

    def mouseMotionAction(self, my_x, y):
        if self.currentClic == self.theSpecialButtonsMapping.MOUSE_LEFT:
            if self.currentMode == MODE_ROTATION:
                self.viewAngle = -(my_x - self.xMouseOnClic) * 0.8
                self.tipAngle = -(y - self.yMouseOnClic) * 0.8
                self.must_rotate = True
            elif self.currentMode == MODE_ZOOM:
                self.zoomLevel *= (1 - 0.005 * (y - self.yMouseOnClic))
            elif self.currentMode == MODE_LIGHT:
                self.lightPosition[0] += (my_x - self.xMouseOnClic) * 100
                self.lightPosition[1] += (y - self.yMouseOnClic) * 100
            else:
                pass
        elif self.currentClic == self.theSpecialButtonsMapping.MOUSE_RIGHT:
            self.centerPosition[0] += (my_x - self.xMouseOnClic) / 1000 * 3
            self.centerPosition[1] -= (y - self.yMouseOnClic) / 1000 * 3
        self.xMouseOnClic = my_x
        self.yMouseOnClic = y

        # GLUT.glutPostRedisplay()

    def keyboardPushAction(self, key):
        # print(chr(key))
        if key == ord(b'+'):
            self.zoomLevel += 0.1 * self.zoomLevel
        elif key == ord(b'-'):
            self.zoomLevel -= 0.1 * self.zoomLevel
        elif key == ord(b'M'):
            self.currentMode += 1
            if self.currentMode >= NUMBER_OF_MODES:
                self.currentMode -= NUMBER_OF_MODES
        # elif key == b'\x1b':
        #         self.stop = True
        # ScaleLength management
        elif key == ord(b'S'):
            self.__displayScale__ = not self.__displayScale__
        elif key == ord(b'>'):
            self.scaleLength += 1
        elif key == ord(b'<'):
            self.scaleLength -= 1
        # Rotation Management
        elif key == self.theSpecialButtonsMapping.KEY_LEFT:
            self.viewAngle = -2
            self.must_rotate = True
        elif key == self.theSpecialButtonsMapping.KEY_RIGHT:
            self.viewAngle = +2
            self.must_rotate = True
        elif key == self.theSpecialButtonsMapping.KEY_UP:
            self.tipAngle = +2
            self.must_rotate = True
        elif key == self.theSpecialButtonsMapping.KEY_DOWN:
            self.tipAngle = -2
            self.must_rotate = True
        elif key == ord(b'X'):
            self.tipAngle = 0
            self.viewAngle = 0
            self.rotation_accumulation = (1, 0, 0, 0)
            self.must_rotate = True
        elif key == ord(b'Y'):
            self.tipAngle = 0
            self.viewAngle = 90
            self.rotation_accumulation = (1, 0, 0, 0)
            self.must_rotate = True
        elif key == ord(b'Z'):
            self.tipAngle = 90
            self.viewAngle = 90
            self.rotation_accumulation = (1, 0, 0, 0)
            self.must_rotate = True
        elif key == ord(b'R'):
            self.__reset__()
        else:
            self.theDeviceDrawer.keyboard_push_action(key)

    def keyboardReleaseAction(self, key, my_x, y):
        pass

    def __draw_axis__(self):
        if self.__displayScale__:
            glColor3f(*self.theDeviceDrawer.get_colour_scalebar())

            glDisable(GL_LIGHTING)
            glDisable(GL_DEPTH_TEST)

            relH = 0.0
            relW = 0.0
            glWindowPos2f(self.VP_MarginW + self.VP_W * (1 + relW) / 2 + 4, self.VP_MarginH + self.VP_H * (1 + relH) / 2 + 4)

            glTranslatef(relW, relH, 0)

            glLineWidth(2.5)
            glBegin(GL_LINES)
            glVertex2f(0.0, 0.0)
            glVertex2f(self.zoomLevel / 1000 * self.scaleLength, 0.0)
            glEnd()

            glLineWidth(1.5)
            glBegin(GL_LINES)
            glVertex2f(0.0, 0.0)
            glVertex2f(0.0, 0.02)
            glVertex2f(self.zoomLevel / 1000 * self.scaleLength, 0.0)
            glVertex2f(self.zoomLevel / 1000 * self.scaleLength, 0.02)
            glEnd()

            # theStr = str(self.scaleLength) + 'mm'
            # for char in theStr:
            #     GLUT.glutBitmapCharacter(GLUT.GLUT_BITMAP_HELVETICA_18, ord(char))
            glEnable(GL_LIGHTING)
            glEnable(GL_DEPTH_TEST)

    def redraw(self):
        if self.theDeviceDrawer is not None:
            glMatrixMode(GL_PROJECTION)
            glLoadIdentity()
            glViewport(self.VP_MarginW, self.VP_MarginH, self.VP_W, self.VP_H)

            # glClearDepth(1)  # just for completeness
            background_color = self.theDeviceDrawer.get_colour_background()
            glClearColor(background_color[0], background_color[1], background_color[2], 1)
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)  # clear the screen
            glMatrixMode(GL_MODELVIEW)
            glLoadIdentity()
            glLightfv(GL_LIGHT0, GL_POSITION, self.lightPosition)

            translationMatrix = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [self.centerPosition[0], self.centerPosition[1], 0, 1]])
            scaleMatrix = np.array([[self.zoomLevel, 0, 0, 0], [0, self.zoomLevel, 0, 0], [0, 0, self.zoomLevel, 0], [0, 0, 0, 1]])
            # Operations on matrix
            if self.must_rotate:
                rot_y = quaternions.normalize(
                    quaternions.axisangle_to_q((1.0, 0.0, 0.0), -self.tipAngle / 180 * math.pi))
                rot_x = quaternions.normalize(
                    quaternions.axisangle_to_q((0.0, 1.0, 0.0), -self.viewAngle / 180 * math.pi))
                self.rotation_accumulation = quaternions.q_mult(self.rotation_accumulation, rot_x)
                self.rotation_accumulation = quaternions.q_mult(self.rotation_accumulation, rot_y)
                self.must_rotate = False

            rotationMatrix = quaternions.q_to_mat4(self.rotation_accumulation)
            transfo = np.matmul(scaleMatrix, np.matmul(rotationMatrix, translationMatrix))
            glLoadMatrixf(transfo)

            if self.theDeviceToDraw is not None:
                self.theDeviceDrawer.draw(self.theDeviceToDraw)
            glLoadIdentity()
            self.__draw_axis__()
            glFlush()

    def get_text_to_write(self):
        lineHeight = 20
        fontSize = 12
        textList = list()
        textList.append(MyText([0, 0, 0], fontSize, 'Press [ESC] to exit', [10, self.VP_H - lineHeight]))
        textList.append(MyText([0, 0, 0], fontSize, 'Reset: [R]', [10, self.VP_H - 2 * lineHeight]))

        textList.append(MyText([0, 0, 0], fontSize, 'Toggle/hide scale: [S], length control: [<] or [>]', [10, 1 * lineHeight]))
        textList.append(MyText([0, 0, 0], fontSize, 'Axis: [X], [Y], [Z]', [10, 2 * lineHeight]))
        return textList

    def __lightingInit__(self):
        glPushMatrix()
        glLoadIdentity()

        glClearColor(self.backgroundColor[0], self.backgroundColor[1], self.backgroundColor[2], 1)
        glShadeModel(GL_SMOOTH)

        glLightfv(GL_LIGHT0, GL_POSITION, self.lightPosition)

        glLightfv(GL_LIGHT0, GL_SPECULAR, [6, 6, 6])
        glLightfv(GL_LIGHT0, GL_AMBIENT, [3, 3, 3])
        glLightfv(GL_LIGHT0, GL_DIFFUSE, [0.6, 0.6, 0.6])

        glLightfv(GL_LIGHT1, GL_POSITION, [22300.0, -55800.0, 0.0, 1.0])
        glLightfv(GL_LIGHT1, GL_SPECULAR, [0.8, 0.8, 0.8])
        # glLightfv(GL_LIGHT1, GL_AMBIENT, [2, 2, 2])
        glLightfv(GL_LIGHT1, GL_DIFFUSE, [0.4, 0.4, 0.4])
        glLightfv(GL_LIGHT1, GL_LINEAR_ATTENUATION, 0)

        glEnable(GL_LIGHTING)
        glEnable(GL_LIGHT0)
        glEnable(GL_LIGHT1)

        glEnable(GL_DEPTH_TEST)
        glEnable(GL_NORMALIZE)
        glPopMatrix()

    def initialize(self):
        self.__lightingInit__()
        self.theDeviceDrawer.get_opengl_options()
        self.tipAngle, self.viewAngle, self.zoomLevel = self.theDeviceDrawer.get_init_camera(self.theDeviceToDraw)

    def __reset__(self):
        deviceDrawer = self.theDeviceDrawer
        deviceToDraw = self.theDeviceToDraw
        keymap = self.theSpecialButtonsMapping
        self.__init__()
        self.set_deviceDrawer(deviceDrawer)
        self.set_deviceToDraw(deviceToDraw)
        self.set_specialButtonsMapping(keymap)




from abc import ABCMeta, abstractmethod


# Proper usage: set numberOfOptimisationParameters first ! then lower/upper then time then objective then compute :)
class AlgorithmInterface(metaclass=ABCMeta):
    """Interface for the optimization algorithm"""

    @abstractmethod
    def initialize(self, initialVectorGuess, listOfOptimizationVariables):
        """
        This function is called once parameters can't be changed anymore, before "get_convergence".

        :param initialVectorGuess: list of variables that describe the initial individual
        :param listOfOptimizationVariables: list of :class:`optimeed.optimize.optiVariable.OptimizationVariable`
        :return:
        """

    @abstractmethod
    def compute(self):  # Launch the optimization
        """
        Launch the optimization

        :return: vector of optimal variables
        """
        pass

    @abstractmethod
    def set_evaluationFunction(self, evaluationFunction, callback_on_evaluation, numberOfObjectives, numberOfConstraints):
        """
        Set the evaluation function and all the necessary callbacks

        :param evaluationFunction: check :meth:`~optimeed.optimize.optimizer.evaluateObjectiveAndConstraints`
        :param callback_on_evaluation: check :meth:`~optimeed.optimize.optimizer.callback_on_evaluation`. Call this function after performing the evaluation of the individuals
        :param numberOfObjectives: int, number of objectives
        :param numberOfConstraints: int, number of constraints
        """
        pass

    @abstractmethod
    def set_maxtime(self, maxTime):  # Maximum time for the optimization
        """Set maximum optimization time (in seconds)"""
        pass

    @abstractmethod
    def get_convergence(self):
        """
        Get the convergence of the optimization

        :return: :class:`~optimeed.optimize.optiAlgorithms.convergence.interfaceConvergence.InterfaceConvergence`
        """
        pass

    def reset(self):
        self.__init__()



from PyQt5 import QtCore, QtGui, QtWidgets
from optimeed.core.ansi2html import Ansi2HTMLConverter


class Widget_text(QtWidgets.QLabel):
    """Widget able to display a text"""
    def __init__(self, theText, is_light=False, convertToHtml=False):
        """

        :param theText: str
        :param is_light: light interface (bool)
        :param convertToHtml: if True: allow to use ansi syntax (= colors)
        """
        super().__init__()

        if is_light:  # Template "Light"
            text_color = QtCore.Qt.black
            text_background = QtCore.Qt.white
        else:
            text_color = QtCore.Qt.white
            text_background = QtCore.Qt.black

        p = self.palette()
        p.setColor(QtGui.QPalette.Background, text_background)
        p.setColor(QtGui.QPalette.WindowText, text_color)
        self.setPalette(p)
        self.setAutoFillBackground(True)

        if convertToHtml:
            conv = Ansi2HTMLConverter()
            self.setTextFormat(QtCore.Qt.RichText)
            self.setText(conv.convert(theText))
        else:
            self.setText(theText)
        self.setFont(QtGui.QFont('Arial', 10))
        self.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)

    def set_text(self, theText, convertToHtml=False):
        """Set the text to display"""
        if convertToHtml:
            conv = Ansi2HTMLConverter()
            self.setTextFormat(QtCore.Qt.RichText)
            self.setText(conv.convert(theText))
        else:
            self.setText(theText)


class Widget_text_scrollable(QtWidgets.QWidget):
    """Same as :class:`~widget_text` but scrollable"""
    def __init__(self, theText, is_light=False, convertToHtml=False):
        super().__init__()
        layout_widget = QtWidgets.QVBoxLayout(self)
        layout_widget.setContentsMargins(0, 0, 0, 0)

        scrollarea = QtWidgets.QScrollArea()
        self.theWidget = Widget_text(theText, is_light, convertToHtml)
        scrollarea.setWidget(self.theWidget)
        scrollarea.setWidgetResizable(True)
        scrollarea.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.layout().addWidget(scrollarea)
        self.setMinimumHeight(150)

    def set_text(self, theText, convertToHtml=False):
        self.theWidget.set_text(theText, convertToHtml)



# Copyright 2015-2018 David Hadka
#
# This file is part of Platypus, a Python module for designing and using
# evolutionary algorithms (EAs) and multiobjective evolutionary algorithms
# (MOEAs).
#
# Platypus is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Platypus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Platypus.  If not, see <http://www.gnu.org/licenses/>.
import pickle
import unittest
import functools
from ..problems import *
from ..operators import RandomGenerator

class TestProblemsSimple(unittest.TestCase):

    def test_DTLZ1(self):
        self._run_test(DTLZ1(2))
    
    def test_DTLZ2(self):
        self._run_test(DTLZ2(2))
        
    def test_DTLZ3(self):
        self._run_test(DTLZ3(2))
        
    def test_DTLZ4(self):
        self._run_test(DTLZ4(2))
        
    def test_DTLZ7(self):
        self._run_test(DTLZ7(2))

    def test_WFG1(self):
        self._run_test(WFG1(2))

    def test_WFG2(self):
        self._run_test(WFG2(2))
        
    def test_WFG3(self):
        self._run_test(WFG3(2))
        
    def test_WFG4(self):
        self._run_test(WFG4(2))
        
    def test_WFG5(self):
        self._run_test(WFG5(2))
        
    def test_WFG6(self):
        self._run_test(WFG6(2))
        
    def test_WFG7(self):
        self._run_test(WFG7(2))
        
    def test_WFG8(self):
        self._run_test(WFG8(2))
        
    def test_WFG9(self):
        self._run_test(WFG9(2))

    def test_UF1(self):
        self._run_test(UF1())
        
    def test_UF2(self):
        self._run_test(UF2())
    
    def test_UF3(self):
        self._run_test(UF3())
        
    def test_UF4(self):
        self._run_test(UF4())
        
    def test_UF5(self):
        self._run_test(UF5())
        
    def test_UF6(self):
        self._run_test(UF6())
        
    def test_UF7(self):
        self._run_test(UF7())
        
    def test_UF8(self):
        self._run_test(UF8())
        
    def test_UF9(self):
        self._run_test(UF9())
        
    def test_UF10(self):
        self._run_test(UF10())
        
    def test_UF11(self):
        self._run_test(UF11())
        
    def test_UF12(self):
        self._run_test(UF12())
        
    def test_UF13(self):
        self._run_test(UF13())
        
    def test_CF1(self):
        self._run_test(CF1())
        
    def test_CF2(self):
        self._run_test(CF2())
        
    def test_CF3(self):
        self._run_test(CF3())
        
    def test_CF4(self):
        self._run_test(CF4())
        
    def test_CF5(self):
        self._run_test(CF5())
        
    def test_CF6(self):
        self._run_test(CF6())
        
    def test_CF7(self):
        self._run_test(CF7())
        
    def test_CF8(self):
        self._run_test(CF8())
        
    def test_CF9(self):
        self._run_test(CF9())
        
    def test_CF10(self):
        self._run_test(CF10())
        
    def test_ZDT1(self):
        self._run_test(ZDT1())
        
    def test_ZDT2(self):
        self._run_test(ZDT2())
        
    def test_ZDT3(self):
        self._run_test(ZDT3())
        
    def test_ZDT4(self):
        self._run_test(ZDT4())
        
    def test_ZDT5(self):
        self._run_test(ZDT5())
        
    def test_ZDT6(self):
        self._run_test(ZDT6())
        
    def _run_test(self, problem):
        if hasattr(problem, "random"):
            solution = problem.random()
        else:
            solution = RandomGenerator().generate(problem)
            
        problem.evaluate(solution)


from .characterization import Characterization
from .mathsToPhysics import MathsToPhysics
from .optiAlgorithms import MultiObjective_GA
from optimeed.core import SaveableObject
from optimeed.core import printIfShown, SHOW_DEBUG
from optimeed.core.commonImport import SHOW_INFO, SHOW_WARNING
import traceback
import time
import copy
import math

default = dict()
default['M2P'] = MathsToPhysics()
default['Charac'] = Characterization()
default['Algo'] = MultiObjective_GA()


class OptimizerSettings(SaveableObject):
    def __init__(self, theDevice, theObjectives, theConstraints, theOptimizationVariables, theOptimizationAlgorithm=None, theMathsToPhysics=None, theCharacterization=None):
        """
        Prepare the optimizer for the optimization.

        :param theDevice: object of type  :class:`~optimeed.core.interfaceDevice.InterfaceDevice`
        :param theObjectives: list of objects of type :class:`~optimeed.optimize.objAndCons.interfaceObjCons.InterfaceObjCons`
        :param theConstraints: list of objects of type :class:`~optimeed.optimize.objAndCons.interfaceObjCons.InterfaceObjCons`
        :param theOptimizationVariables: list of objects of type :class:`~optimeed.optimize.optiVariable.OptimizationVariable`
        :param theOptimizationAlgorithm: list of objects of type :class:`~optimeed.optimize.optiAlgorithms.algorithmInterface.AlgorithmInterface`
        :param theMathsToPhysics: object of type :class:`~optimeed.optimize.mathsToPhysics.interfaceMathsToPhysics.InterfaceMathsToPhysics`
        :param theCharacterization: object of type :class:`~optimeed.optimize.characterization.interfaceCharacterization.InterfaceCharacterization`
        :return:
        """
        self.theDevice = theDevice
        self.theMathsToPhysics = theMathsToPhysics if theMathsToPhysics is not None else default['M2P']
        self.theCharacterization = theCharacterization if theCharacterization is not None else default['Charac']
        self.theObjectives = theObjectives
        self.theConstraints = theConstraints
        self.theOptimizationAlgorithm = theOptimizationAlgorithm if theOptimizationAlgorithm is not None else default['Algo']
        self.theOptimizationVariables = theOptimizationVariables

    def get_additional_attributes_to_save(self):
        return ["theDevice", "theMathsToPhysics", "theCharacterization", "theOptimizationAlgorithm"]

    def get_additional_attributes_to_save_list(self):
        return ["theObjectives", "theConstraints", "theOptimizationVariables"]

    def get_device(self):
        return self.theDevice

    def get_M2P(self):
        return self.theMathsToPhysics

    def get_charac(self):
        return self.theCharacterization

    def get_optivariables(self):
        return self.theOptimizationVariables

    def get_objectives(self):
        return self.theObjectives
    
    def get_constraints(self):
        return self.theConstraints

    def get_optialgorithm(self):
        return self.theOptimizationAlgorithm
    
    
class _Evaluator:
    """This is the main class that serves as evaluator. This class is NOT process safe (i.e., copy of it might be generated upon process call)"""

    def __init__(self, optimization_parameters: OptimizerSettings):
        # Variables defining an optimization problem
        self.theDevice = optimization_parameters.get_device()
        self.theMathsToPhysics = optimization_parameters.get_M2P()
        self.theCharacterization = optimization_parameters.get_charac()
        self.theObjectives = optimization_parameters.get_objectives()
        self.theConstraints = optimization_parameters.get_constraints()
        self.theOptimizationVariables = optimization_parameters.get_optivariables()
        self.startingTime = None

    def start(self):
        self.startingTime = time.time()

    def evaluate(self, x):
        """
        Evaluates the performances of device associated to entrance vector x. Outputs the objective function and the constraints,
        and other data used in optiHistoric.

        This function is NOT process safe: "self." is a FORK in multiprocessing algorithms.
        It means that the motor originally contained in self. is modified only in the fork, and only gathered by reaching the end of the fork.

        :param x: Input mathematical vector from optimization algorithm
        :return: dictionary, containing objective values (list of scalar), constraint values (list of scalar), and other info (motor, time)
        """
        copyDevice = copy.copy(self.theDevice)
        self.theMathsToPhysics.fromMathsToPhys(x, copyDevice, self.theOptimizationVariables)

        characterization_failed = False
        # noinspection PyBroadException
        try:
            self.theCharacterization.compute(copyDevice)
        except Exception:
            characterization_failed = True
            printIfShown("An error in characterization. Set objectives to inf. Error :" + traceback.format_exc(), SHOW_WARNING)

        nbr_of_objectives = len(self.theObjectives)
        objective_values = [float('inf')]*nbr_of_objectives

        nbr_of_constraints = len(self.theConstraints)
        constraint_values = [float('inf')] * nbr_of_constraints

        if not characterization_failed:
            for i in range(nbr_of_objectives):
                # noinspection PyBroadException
                try:
                    objective_values[i] = self.theObjectives[i].compute(copyDevice)
                    if math.isnan(objective_values[i]):
                        objective_values[i] = float('inf')
                except Exception:
                    objective_values[i] = float('inf')
                    printIfShown("An error in objectives. inf value has been set to continue execution. Error:" + traceback.format_exc(), SHOW_DEBUG)

            for i in range(nbr_of_constraints):
                # noinspection PyBroadException
                try:
                    constraint_values[i] = self.theConstraints[i].compute(copyDevice)
                except Exception:
                    constraint_values[i] = float('inf')
                    printIfShown("An error in constraints. NaN value has been set to continue execution. Error:" + traceback.format_exc(), SHOW_DEBUG)

        valuesToReturn = dict()
        valuesToReturn["params"] = x
        valuesToReturn["device"] = copyDevice
        valuesToReturn["time"] = time.time()-self.startingTime
        valuesToReturn["objectives"] = objective_values
        valuesToReturn["constraints"] = constraint_values
        return valuesToReturn  # objective_values, constraint_values

    def reevaluate_solutions(self, x_solutions):
        resultsdevices = [None] * len(x_solutions)
        for i, x_solution in enumerate(x_solutions):
            self.theMathsToPhysics.fromMathsToPhys(x_solution, self.theDevice, self.theOptimizationVariables)
            self.theCharacterization.compute(self.theDevice)
            resultsdevices[i] = copy.deepcopy(self.theDevice)
        return resultsdevices


def run_optimization(optimization_parameters: OptimizerSettings, opti_historic, max_opti_time_sec=10):
    """
    Perform the optimization.

    :return: list of the best optimized devices, convergence information
    """

    evaluator = _Evaluator(optimization_parameters)
    theOptimizationAlgorithm = optimization_parameters.get_optialgorithm()

    # Initialize opti algorithms
    opti_historic.start(optimization_parameters)
    evaluator.start()

    initialVectorGuess = evaluator.theMathsToPhysics.fromPhysToMaths(evaluator.theDevice, evaluator.theOptimizationVariables)

    theOptimizationAlgorithm.set_maxtime(max_opti_time_sec)
    theOptimizationAlgorithm.set_evaluationFunction(evaluator.evaluate, opti_historic.log_after_evaluation, len(evaluator.theObjectives), len(evaluator.theConstraints))
    # Initialize the algorithm
    theOptimizationAlgorithm.initialize(initialVectorGuess, evaluator.theOptimizationVariables)

    # Get track of convergence
    convergence = theOptimizationAlgorithm.get_convergence()
    opti_historic.set_convergence(convergence)

    # Start optimization
    printIfShown("Performing optimization", SHOW_INFO)
    x_solutions = theOptimizationAlgorithm.compute()
    printIfShown("Optimization ended", SHOW_INFO)

    # Manage results
    success, best_devices = opti_historic.get_best_devices_without_reevaluating(x_solutions)
    if not success:
        printIfShown("Could not retrieve best devices from database ... Reevaluating", SHOW_INFO)
        best_devices = evaluator.reevaluate_solutions(x_solutions)

    # Set results and save
    opti_historic.set_results(best_devices)
    opti_historic.save()

    return best_devices, convergence



from optimeed.core.linkDataGraph import LinkDataGraph, HowToPlotGraph
from optimeed.core import Option_class, Option_bool
from optimeed.optimize.optimizer import run_optimization
from optimeed.visualize.mainWindow import MainWindow
from optimeed.visualize.graphs.widget_graphsVisual import Widget_graphsVisual
from optimeed.visualize.process_mainloop import start_qt_mainloop, stop_qt_mainloop, process_qt_events
from optimeed.visualize.onclick import Onclick_representDevice, Represent_brut_attributes
from threading import Thread
from threading import Timer
from PyQt5 import QtCore
import queue
import os
import traceback


def check_if_must_plot(elem):
    constraints = elem.constraints
    for constraint in constraints:
        if constraint > 0:
            return False
    return True


def run_optimization_displayer(*args, **kwargs):
    # Wrapper class to work in a thread
    resultsOpti, convergence = run_optimization(*args, **kwargs)
    stop_qt_mainloop()
    return resultsOpti, convergence


class OptimizationDisplayer(Option_class):
    """Class used to display optimization process in real time"""
    signal_optimization_over = QtCore.pyqtSignal()

    SHOW_CONSTRAINTS = 0

    def __init__(self, theOptiParameters, theOptiHistoric, additionalWidgets=None, light_background=False):
        """

        :param theOptiParameters: list of :class:`~optimeed.optimize.optimizer.OptimizationParameters`
        :param theOptiHistoric: :class:`~optimeed.optimize.optiHistoric.OptiHistoric` (or similar)
        :param additionalWidgets: list of QtWidgets (instantiated)
        :param light_background: boolean, True or False for White or Black background color in graphs
        """
        super().__init__()
        self.add_option(self.SHOW_CONSTRAINTS, Option_bool("Display out of constraint elements", True))
        self.theOptiHistoric = theOptiHistoric
        self.listOfObjectives = theOptiParameters.get_objectives()
        if additionalWidgets is None:
            self.additionalWidgets = list()
        else:
            self.additionalWidgets = additionalWidgets

        self.theActionsOnClick = None
        self.theDataLink = None
        self.myWidgetGraphsVisuals = None
        self.id_logOpti = None
        self.windowOpti = None
        self.windowConvergence = None
        self.timer_autorefresh = None
        self.light_background = light_background

    def set_actionsOnClick(self, theList):
        """Set actions to perform on click, list of :class:`~optimeed.visualize.gui.widgets.widget_graphs_visual.on_graph_click_interface`"""
        self.theActionsOnClick = theList

    def generate_optimizationGraphs(self):
        """Generates the optimization graphs.
       :return: :class:`~optimeed.core.graphs.Graphs`, :class:`~optimeed.core.linkDataGraph.LinkDataGraph`, :class:'~optimeed.visulaize.gui.widgets.widget_graphs_visual.widget_graphs_visual"""

        collection_devices = self.theOptiHistoric.get_devices()
        collection_logOpti = self.theOptiHistoric.get_logopti()

        theDataLink = LinkDataGraph()
        id_logOpti = theDataLink.add_collection(collection_logOpti)
        theDataLink.set_shadow_collection(id_logOpti, collection_devices)  # Link the 2 collections

        check_plot = None if self.get_option_value(self.SHOW_CONSTRAINTS) else check_if_must_plot

        if len(self.listOfObjectives) == 2:
            x_label = self.listOfObjectives[0].get_name()
            y_label = self.listOfObjectives[1].get_name()
            howToPlot = HowToPlotGraph('objectives[0]', 'objectives[1]', {'x_label': x_label, 'y_label': y_label, 'is_scattered': True}, check_if_plot_elem=check_plot)
            theDataLink.add_graph(howToPlot)
        elif len(self.listOfObjectives) == 3:
            for i, objective in enumerate(self.listOfObjectives):
                i_plus_1 = i+1 if i != 2 else 2
                curr_i = i if i != 2 else 0
                x_label = self.listOfObjectives[curr_i].get_name()
                y_label = self.listOfObjectives[i_plus_1].get_name()
                howToPlot = HowToPlotGraph('objectives[{}]'.format(curr_i), 'objectives[{}]'.format(i_plus_1), {'x_label': x_label, 'y_label': y_label, 'is_scattered': True}, check_if_plot_elem=check_plot)
                theDataLink.add_graph(howToPlot)
        else:
            for i, objective in enumerate(self.listOfObjectives):
                x_label = 'Time [s]'
                y_label = self.listOfObjectives[i].get_name()
                howToPlot = HowToPlotGraph('time', 'objectives[{}]'.format(i), {'x_label': x_label, 'y_label': y_label, 'is_scattered': True}, check_if_plot_elem=check_plot)
                theDataLink.add_graph(howToPlot)

        theGraphs = theDataLink.get_graphs()
        self.theDataLink = theDataLink
        self.id_logOpti = id_logOpti

        """Generate live graph for the optimization"""
        # Set actions on graph click ...
        if self.theActionsOnClick is None:
            self.theActionsOnClick = list()
            self.theActionsOnClick.append(Onclick_representDevice(theDataLink, [Represent_brut_attributes()]))

        self.myWidgetGraphsVisuals = Widget_graphsVisual(theGraphs, highlight_last=True, refresh_time=-1, is_light=self.light_background)
        self.__set_graphs_disposition()
        return theGraphs, theDataLink, Widget_graphsVisual

    def __change_appearance_violate_constraints(self):
        graphs, theCol = self.theDataLink.get_graph_and_trace_from_idCollection(self.id_logOpti)
        all_constraints = theCol.get_list_attributes("constraints")
        violate_constraint_indices = list()
        for index, constraints in enumerate(all_constraints):
            for constraint in constraints:
                if constraint > 0:
                    violate_constraint_indices.append(index)  # Index in data
        for graph, trace in graphs:
            theTrace = self.myWidgetGraphsVisuals.get_graph(graph).get_trace(trace)
            theTrace.set_brushes(violate_constraint_indices, (250, 0, 0))

    def __refresh(self):
        self.theDataLink.update_graphs()
        if self.get_option_value(self.SHOW_CONSTRAINTS):
            self.__change_appearance_violate_constraints()
        self.myWidgetGraphsVisuals.update_graphs()

    def start_autorefresh(self, timer_autosave):
        self.__refresh()
        if timer_autosave > 0:
            self.timer_autorefresh = Timer(timer_autosave, lambda: self.start_autorefresh(timer_autosave))
            self.timer_autorefresh.daemon = True
            self.timer_autorefresh.start()

    def stop_autorefresh(self):
        if self.timer_autorefresh is not None:
            self.timer_autorefresh.cancel()

    def __set_graphs_disposition(self):
        """Set nicely the graphs disposition"""
        allGraphs = self.myWidgetGraphsVisuals.get_all_graphsVisual()
        row = 1
        col = 1
        maxNbrRow = 1
        currRow = 1
        for idGraph in allGraphs:
            self.myWidgetGraphsVisuals.set_graph_disposition(idGraph, row, col)
            if currRow >= maxNbrRow:
                row = 1
                col += 1
            else:
                row += 1
            currRow = row

    def launch_optimization(self, args_opti, kwargs_opti, refresh_time=0.1, max_nb_points_convergence=100):
        """Perform the optimization and spawn the convergence graphs afterwards.
        :param args_opti: arguments (as list) destined to launch the optimization
        :param kwargs_opti: keywords arguments (as dict) destined to launch the optimization
        :param refresh_time: float indicating the refresh time of the graphs. If it becomes laggy -> use a higher one.
        :param max_nb_points_convergence: maximum number of points in the graph that displays the convergence. Put None if performance is not an issue."""

        self.create_main_window()  # Create the gui

        # Launch optimization in a separate thread
        self.start_autorefresh(refresh_time)

        que = queue.Queue()  # Put results in queue
        threadOpti = Thread(target=lambda: que.put(run_optimization_displayer(*args_opti, **kwargs_opti)))
        threadOpti.start()

        start_qt_mainloop()  # Automatically stops at the end of the optimization

        threadOpti.join()  # This means stop_qt_mainloop has been called (opti finished)
        self.stop_autorefresh()
        resultsOpti, convergence = que.get()

        # noinspection PyBroadException
        try:
            self.windowConvergence, widgetGraphsVisualConvergence = self.display_graphs(convergence.get_graphs(max_number_of_points=max_nb_points_convergence))
            process_qt_events()

            # Save graphs:
            path_results = self.theOptiHistoric.foldername
            path_results = os.path.join(path_results, "results")
            os.makedirs(path_results)
            self.myWidgetGraphsVisuals.exportGraphs(os.path.join(path_results, "logopti"))
            widgetGraphsVisualConvergence.exportGraphs(os.path.join(path_results, "convergence"))
        except Exception:
            traceback.print_exc()

        return resultsOpti, convergence

    def close_windows(self):
        if self.windowOpti is not None:
            self.windowOpti.close()

        if self.windowConvergence is not None:
            self.windowConvergence.close()

    def display_graphs(self, theGraphs):
        myWidgetGraphsVisuals = Widget_graphsVisual(theGraphs, highlight_last=False, refresh_time=-1, is_light=self.light_background)
        myWindow = MainWindow([myWidgetGraphsVisuals], actionOnWindowClosed=lambda *args: stop_qt_mainloop())
        if not theGraphs.is_empty():
            myWindow.run(False)
        return myWindow, myWidgetGraphsVisuals

    def create_main_window(self):
        """From the widgets and the actions on click, spawn a window and put a gui around widgetsGraphsVisual."""
        self.myWidgetGraphsVisuals.set_actions_on_click(self.theActionsOnClick)
        # Add additional widgets to the visualisation
        listOfWidgets = self.additionalWidgets
        listOfWidgets.append(self.myWidgetGraphsVisuals)
        myWindow = MainWindow(listOfWidgets, actionOnWindowClosed=lambda *args: stop_qt_mainloop())
        myWindow.set_actionOnClose(lambda _, __: self.stop_autorefresh())
        myWindow.run(False)
        self.windowOpti = myWindow



from optimeed.core import LinkDataGraph, HowToPlotGraph, merge_two_dicts
from optimeed.visualize.graphs.widget_graphsVisual import Widget_graphsVisual
from optimeed.visualize.mainWindow import MainWindow
from optimeed.core import ListDataStruct, Performance_ListDataStruct, ListDataStruct_Interface
from optimeed.core import Graphs
import os


class _OptiProjectLoader:
    """A loader for an opti project."""
    def __init__(self, foldername, kwargsPlot=None):
        """

        :param foldername: the folder containing the saved files.
        :param kwargsPlot: Check kgwargs `~optimeed.core.graphs.Data`
        """
        self.logopti = ListDataStruct.load(os.path.join(foldername, "logopti.json"))
        try:
            self.theDevices = Performance_ListDataStruct.load(os.path.join(foldername, "autosaved.json"))
        except KeyError:  # If it fails, it is probably a ListDataStruct
            self.theDevices = ListDataStruct.load(os.path.join(foldername, "autosaved.json"))
        self.theConvergence = ListDataStruct.load(os.path.join(foldername, "optiConvergence.json"))
        self.kwargsPlot = kwargsPlot

    def get_devices(self) -> ListDataStruct_Interface:
        return self.theDevices

    def get_logopti(self) -> ListDataStruct_Interface:
        return self.logopti

    def get_convergence(self):
        return self.theConvergence.get_data_at_index(0)

    def get_kwargs(self):
        return dict() if self.kwargsPlot is None else self.kwargsPlot

    def get_nbr_objectives(self):
        return len(self.logopti.get_data_at_index(0).objectives)


class ViewOptimizationResults:
    """Convenience class to display the results of an optimization"""

    def __init__(self):
        self.optiProjects = list()
        self.theDataLink = LinkDataGraph()

    def add_opti_project(self, foldername, kwargsPlot=None):
        """Add an opti project to visualize.

        :param foldername: the folder containing the saved files. (as string)
        :param kwargsPlot: Check kgwargs `~optimeed.core.graphs.Data`
        """
        self.optiProjects.append(_OptiProjectLoader(foldername, kwargsPlot))

    def get_data_link(self) -> LinkDataGraph:
        """Return the object :class:`~optimeed.core.linkDataGraph.LinkDataGraph`"""
        return self.theDataLink

    def display_graphs(self, theActionsOnClick=None, kwargs_common=None, keep_alive=True, max_nb_points_convergence=None, light_background=False):
        """Generates the optimization graphs.

        :param theActionsOnClick: list of actions to perform when a graph is clicked
        :param kwargs_common: plot options (from Data class) to apply to all the graphs (ex: {"is_scattered": True}).
        :param keep_alive: if set to true, this method will be blocking. Otherwise you should manually call start_qt_mainloop().
        :param max_nb_points_convergence: maximum number of points in the graph that displays the convergence. Put None if performance is not an issue.
        :param light_background: boolean, True or False for White or Black background color in graphs
        :return: widget_graphs_visual for the log opti, widget_graphs_visual for the convergence (:class:`~widget_graphs_visual`)
        """
        if theActionsOnClick is None:
            theActionsOnClick = list()

        nbrObjectives = self.optiProjects[0].get_nbr_objectives()
        theDataLink = self.theDataLink

        # Creates empty graphs
        listOf_howtoplot = list()

        base_kwargs = {'is_scattered': True}
        if kwargs_common is not None:
            base_kwargs.update(kwargs_common)

        if nbrObjectives == 2:  # bi-objective -> each objective is an axis
            howToPlot = HowToPlotGraph('objectives[0]', 'objectives[1]', base_kwargs)
            theDataLink.add_graph(howToPlot)  # Creates a graph
            listOf_howtoplot.append(howToPlot)
        elif nbrObjectives == 3:  # 3-objectives -> plot graphs by pairs of objective
            for i in range(nbrObjectives):
                i_plus_1 = i + 1 if i != 2 else 2
                curr_i = i if i != 2 else 0
                howToPlot = HowToPlotGraph('objectives[{}]'.format(curr_i), 'objectives[{}]'.format(i_plus_1), base_kwargs)
                theDataLink.add_graph(howToPlot)  # Creates a graph
                listOf_howtoplot.append(howToPlot)
        else:  # Plot each objective as a function of the time
            for i in range(nbrObjectives):
                howToPlot = HowToPlotGraph(None, 'objectives[{}]'.format(i), base_kwargs)
                theDataLink.add_graph(howToPlot)  # Creates a graph
                listOf_howtoplot.append(howToPlot)

        # Graphs are created, now add the data to it.
        for theOptiProject in self.optiProjects:
            collection_devices = theOptiProject.get_devices()
            collection_logOpti = theOptiProject.get_logopti()

            # There is a bit of processing here to split the full data set into two:
            # Those that respects the constraints and those that do not
            all_constraints = collection_logOpti.get_list_attributes("constraints")
            violate_constraint_indices = list()
            max_index = min(collection_devices.get_nbr_elements(), collection_logOpti.get_nbr_elements())

            for index, constraints in enumerate(all_constraints):
                if index > max_index:
                    break
                for constraint in constraints:
                    if constraint > 0:
                        violate_constraint_indices.append(index)  # Index in data

            # Based on that -> create subsets
            subsets_opti = list()  # will contain master, slave, kwargs

            if len(violate_constraint_indices):
                all_indices = set(range(max_index))
                indices_ko = violate_constraint_indices
                indices_ok = [u for u in all_indices if u not in set(indices_ko)]

                collection_devices_ok = collection_devices.extract_collection_from_indices(indices_ok)
                collection_logOpti_ok = collection_logOpti.extract_collection_from_indices(indices_ok)
                subsets_opti.append((collection_devices_ok, collection_logOpti_ok, {"legend": "Respects constraints"}))

                collection_devices_ko = collection_devices.extract_collection_from_indices(indices_ko)
                collection_logOpti_ko = collection_logOpti.extract_collection_from_indices(indices_ko)
                subsets_opti.append((collection_devices_ko, collection_logOpti_ko, {"legend": "Violates constraints"}))
            else:
                subsets_opti.append((collection_devices, collection_logOpti, {}))

            # We add them
            for collection_devices, collection_logOpti, newKwargs in subsets_opti:
                id_logOpti = theDataLink.add_collection(collection_logOpti, kwargs=merge_two_dicts(theOptiProject.get_kwargs(), newKwargs))
                """The trick here is that the objective functions is not directly stocked in collection_devices but in collection_logOpti. 
                So we display the objectives coming from collection_logOpti but we link collection_devices from it.
                So that when a point is clicked, the action is performed on the device and not on the logOpti."""
                theDataLink.set_shadow_collection(id_logOpti, collection_devices)

        """Create the widget of the graphs, and the associated GUI"""
        myWidgetGraphsVisuals = Widget_graphsVisual(theDataLink.get_graphs(), actionsOnClick=theActionsOnClick, highlight_last=False, refresh_time=-1, is_light=light_background)

        """Spawn the GUI"""
        theWindow = MainWindow([myWidgetGraphsVisuals])
        theWindow.run(hold=False)

        """Spawn an other GUI for the convergence graphs"""
        graphs_convergence = Graphs()
        for theOptiProject in self.optiProjects:
            newGraphs = theOptiProject.get_convergence().get_graphs(max_number_of_points=max_nb_points_convergence)
            newGraphs.get_first_graph().get_trace(0).set_legend(theOptiProject.get_kwargs().get("legend", ''))
            graphs_convergence.merge(newGraphs)
        wg_graphs_convergence = Widget_graphsVisual(graphs_convergence, highlight_last=False, refresh_time=-1, is_light=light_background)

        myWindow = MainWindow([wg_graphs_convergence])
        myWindow.run(hold=keep_alive)

        return myWidgetGraphsVisuals, wg_graphs_convergence



from setuptools import setup, find_namespace_packages
from optimeed import VERSION
with open("README.md", "r") as fh:
    long_description = fh.read()

extras = {
    'with_matplotlib': ['matplotlib'],  # Hold CTRL to select multiple points on graphs
    'with_pandas': ['pandas'],  # Export XLS the collection
    'with_plotly': ['plotly']  # 3D plots
}

setup(
    name="optimeed",
    version="{}".format(VERSION),
    author="Christophe De Greef",
    author_email="christophe.degreef@uclouvain.be",
    description="Powerful optimization and vizualisation tool.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.immc.ucl.ac.be/chdegreef/optimeed",
    packages=find_namespace_packages(include=['optimeed.*']),
    include_package_data=True,
    install_requires=['numpy', 'PyOpenGL', 'PyQt5', 'pytypes', 'SALib'],
    extras_require=extras,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Topic :: Scientific/Engineering"
    ],
    python_requires='>=3.6',
)



from .onclickInterface import OnclickInterface
from optimeed.visualize.graphs.pyqtgraph import mkPen, TextItem, GraphicsObject
import numpy as np
from PyQt5 import QtGui, QtCore


class _LineItem(GraphicsObject):
    def __init__(self, point1, point2):
        GraphicsObject.__init__(self)
        self.point1 = point1
        self.point2 = point2

    def paint(self, p, *args):
        p.setRenderHint(QtGui.QPainter.Antialiasing)
        pen = mkPen("b")
        pen.setWidthF(pen.widthF())
        p.setPen(pen)
        p.drawLine(self.point1, self.point2)

        # p.setFont(QtGui.QFont('Helvetica', 1))
        # p.drawText(self.point1, self.text)  # Problem

    def boundingRect(self):
        vr = self.viewRect()  # bounds of containing ViewBox mapped to local coords.
        if vr is None:
            return QtCore.QRectF()
        return QtCore.QRectF(self.point1, self.point2)


class Onclick_measure(OnclickInterface):
    """On Click: Measure distance. Click on two points to perform that action"""

    def __init__(self):
        self.point1 = None
        self.point2 = None
        self.theGraphVisual = None
        self.clicked_graph = None
        self.added_items = list()
        self.gtp1 = None
        self.gtp2 = None

    def graph_clicked(self, the_graph_visual, index_graph, index_trace, indices_points):
        if self.clicked_graph is None:
            self.clicked_graph = index_graph
            self.theGraphVisual = the_graph_visual

        if self.clicked_graph == index_graph:
            theTrace = the_graph_visual.get_graph(index_graph).get_trace(index_trace)
            point = indices_points[0]
            clickedPoint = theTrace.get_point(point).pos()
            if self.point1 is None:
                self.point1 = clickedPoint
                self.gtp1 = (index_graph, index_trace, point)
            elif self.point2 is None:
                self.point2 = clickedPoint
                self.display_distance()
                self.gtp2 = (index_graph, index_trace, point)
            else:
                self.point1 = clickedPoint
                self.point2 = None
                self.reset_distance()
                self.gtp1 = (index_graph, index_trace, point)
            theTrace.set_brush(point, (250, 250, 0))

    def reset_distance(self):
        for item in self.added_items:
            self.theGraphVisual.get_graph(self.clicked_graph).remove_feature(item)
        self.added_items = list()
        for gtp in [self.gtp1, self.gtp2]:
            if len(gtp):
                self.theGraphVisual.get_graph(gtp[0]).get_trace(gtp[1]).reset_brush(gtp[2])
        self.gtp1 = list()
        self.gtp2 = list()

    def display_distance(self):
        distance = np.sqrt((self.point1.x()-self.point2.x())**2 + (self.point1.y()-self.point2.y())**2)
        lineItem = _LineItem(self.point1, self.point2)
        textItem = TextItem("{:#.4g}".format(distance), color="b", anchor=(0.5, 0.5))
        textItem.setPos((self.point2 + self.point1)/2.0)
        # textItem.setParentItem(lineItem)
        # textItem.setAngle(np.arctan((self.point2.y()-self.point1.y())/(self.point2.x()-self.point1.x()))*180/np.pi)
        self.theGraphVisual.get_graph(self.clicked_graph).add_feature(lineItem)
        self.theGraphVisual.get_graph(self.clicked_graph).add_feature(textItem)
        self.added_items.append(lineItem)
        self.added_items.append(textItem)

    def get_name(self):
        return "Measure distance"




from PyQt5 import QtWidgets, uic
from optimeed.core.tools import rgetattr, get_recursive_attrs
from optimeed.core import LinkDataGraph, HowToPlotGraph
from optimeed.core import SHOW_ERROR, printIfShown
from optimeed.visualize.selector import Onselect_highlight, Onselect_newTrace, Onselect_splitTrace
from optimeed.visualize.graphs import Widget_graphsVisual
from optimeed.visualize.mainWindow import MainWindow
from optimeed.visualize.process_mainloop import start_qt_mainloop
import traceback
import os


def _is_object_selected(object_in, min_max_attributes):
    for name_attr, min_attr, max_attr in min_max_attributes:
        value = rgetattr(object_in, name_attr)
        if value is None:
            return False
        type_init = type(value)

        try:
            min_attr = type_init(min_attr)
            has_min = True
        except (TypeError, ValueError):
            has_min = False
        try:
            max_attr = type_init(max_attr)
            has_max = True
        except (TypeError, ValueError):
            has_max = False

        if not (has_min or has_max):
            return False
        if not has_min:
            if rgetattr(object_in, name_attr) > type_init(max_attr):
                return False
        if not has_max:
            if rgetattr(object_in, name_attr) < type_init(min_attr):
                return False
        if not (min_attr <= value <= max_attr or max_attr <= value <= min_attr):
            return False
    return True


def _select_and_apply_action(theCollections, min_max_attributes, theAction, selectionName):
    id_selections = list()
    try:

        for collection in theCollections:
            selected_data_indices = []
            unselected_data_indices = []

            for k, data in enumerate(collection.get_data_generator()):
                if _is_object_selected(data, min_max_attributes):
                    selected_data_indices.append(k)
                else:
                    unselected_data_indices.append(k)
            id_selections.append(theAction.selector_updated(selectionName, collection, selected_data_indices, unselected_data_indices))
    except KeyboardInterrupt:
        raise
    except Exception:
        printIfShown("{}".format(traceback.format_exc()), SHOW_ERROR)
    return id_selections


class CollectionDisplayer(QtWidgets.QMainWindow):
    """GUI to display a collection."""

    def __init__(self):
        super().__init__()  #
        uic.loadUi(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'displayCollections_gui.ui'), self)

        # Structures
        self.howToPlot = HowToPlotGraph('', '', kwargs_graph={"is_scattered": True, "alpha": 255}, meta="")
        self.theDataLink = LinkDataGraph()
        self.theDataLink.add_graph(self.howToPlot)
        self.theWGPlot = Widget_graphsVisual(self.theDataLink.get_graphs(), refresh_time=-1)
        self.window_plot = MainWindow([self.theWGPlot])

        # Fill in
        possible_actions = list()
        possible_actions.append(Onselect_highlight(self.theDataLink, self.theWGPlot))
        possible_actions.append(Onselect_newTrace(self.theDataLink))
        possible_actions.append(Onselect_splitTrace(self.theDataLink))
        for action in possible_actions:
            self.selector_action.addItem(action.get_name())
        self.selector_action.setCurrentIndex(0)
        self.possible_actions = possible_actions
        self.curr_action = possible_actions[0]
        # Connect
        self.selector_action.currentIndexChanged.connect(lambda index: self.set_action_selector(self.possible_actions[index]))
        self.set_X.clicked.connect(self._set_x)
        self.set_Y.clicked.connect(self._set_y)
        self.set_Z.clicked.connect(self._set_z)

        self.selector_to.clicked.connect(self._selector_to)
        self.selector_cancel.clicked.connect(self._remove_item_selector)
        self.reset_colors.clicked.connect(self._reset_colors)
        self.items_selected.cellChanged.connect(self._apply_selector)
        self.previous_selection = list()

        # SHOW
        self.show()
        self.window_plot.show()

        self.initialized = False

    # Users functions
    def add_collection(self, theCollection, name=""):
        """Add a collection to the GUI"""
        theIdCollection = self.theDataLink.add_collection(theCollection, {"legend": name})
        if not self.initialized:
            self._initialize(theCollection)
        return theIdCollection

    def set_shadow(self, master_collectionId, shadow_collection):
        """Set a shadow collection to master_collectionID (see DataLink.set_shadow_collection)"""
        self.theDataLink.set_shadow_collection(master_collectionId, shadow_collection)

    def remove_collection(self, theCollection):
        """Remove collection from the GUI"""
        self.theDataLink.remove_collection(self.theDataLink.get_idcollection_from_collection(theCollection))

    def update_graphs(self):
        self.theDataLink.update_graphs()

    def set_actions_on_click(self, theActionsOnClick):
        """Set actions to be performed when graph is clicked"""
        self.theWGPlot.set_actions_on_click(theActionsOnClick)

    @staticmethod
    def run():
        start_qt_mainloop()

    def get_datalink(self):
        return self.theDataLink

    # Others

    def _initialize(self, theCollection):
        first_data = theCollection.get_data_at_index(0)
        all_attrs = get_recursive_attrs(first_data)

        self.available_plot_data.set_list(all_attrs)
        self.items_to_select.set_entries(all_attrs)
        self.items_selected.set_entries(all_attrs, hidden=True)

        self.items_selected.myTableWidget.setHorizontalHeaderLabels(["Attribute", "Min", "Max"])
        self.items_to_select.myTableWidget.setHorizontalHeaderLabels(["Attribute", "Min", "Max"])

        self.initialized = True

    # Plot methods

    def _set_x(self):
        name = self.available_plot_data.get_name_selected()
        self.howToPlot.kwargs_graph.update({"x_label": name})
        if name is not None:
            howToPlot = self.theDataLink.get_howToPlotGraph(0)
            howToPlot.attribute_x = name
        self.update_graphs()

    def _set_y(self):
        name = self.available_plot_data.get_name_selected()
        self.howToPlot.kwargs_graph.update({"y_label": name})

        if name is not None:
            howToPlot = self.theDataLink.get_howToPlotGraph(0)
            howToPlot.attribute_y = name
        self.update_graphs()

    def _set_z(self):
        name = self.available_plot_data.get_name_selected()
        howToPlot = self.theDataLink.get_howToPlotGraph(0)

        if name is not None:
            howToPlot.meta = name
        else:
            howToPlot.meta = None
        self.update_graphs()

    # Selector methods

    def set_action_selector(self, theAction):
        self.curr_action = theAction

    def _selector_to(self):
        results, rows = self.items_to_select.get_entries_selected()
        if len(rows):
            [self.items_to_select.force_hide_row(row) for row in rows]
            [self.items_selected.remove_forced_hide_row(row) for row in rows]
            self.items_selected.blockSignals(True)
            for k, row in enumerate(rows):
                for col in [1, 2]:
                    self.items_selected.set_item(row, col, results[k][col])
            self.items_selected.blockSignals(False)

            self._apply_selector()

    def _remove_item_selector(self):
        results, rows = self.items_selected.get_entries_selected()
        if len(rows):
            [self.items_selected.force_hide_row(row) for row in rows]
            [self.items_to_select.remove_forced_hide_row(row) for row in rows]

            for k, row in enumerate(rows):
                for col in [1, 2]:
                    self.items_to_select.set_item(row, col, results[k][col])
            self._apply_selector()

    def _cancel_selector(self):
        # Automatically called by _apply_selector

        # noinspection PyBroadException
        try:
            if self.previous_selection:
                for graphId, traceId, id_selections in self.previous_selection:
                    for id_selection in id_selections:
                        self.curr_action.cancel_selector(id_selection)
                self.previous_selection = list()
        except Exception:
            pass

    def _apply_selector(self):
        if self.initialized:
            self._cancel_selector()

            for graphId in self.theDataLink.get_idGraphs():
                for traceId in self.theDataLink.get_idTraces(graphId):
                    selected, _ = self.items_selected.get_shown_entries()
                    if selected:
                        legend = self.theDataLink.kwargs_collections[self.theDataLink.get_idCollection_from_graph(graphId, traceId)].get('legend', '')  # Tricky line, might need to update DataLink
                        id_selections = _select_and_apply_action([self.theDataLink.get_collection_from_graph(graphId, traceId, getShadow=False)], selected, self.curr_action, legend)
                        self.previous_selection.append([graphId, traceId, id_selections])
            self.update_graphs()

    def _reset_colors(self):
        for graphId in self.theDataLink.get_idGraphs():
            for traceId in self.theDataLink.get_idTraces(graphId):
                self.theWGPlot.get_trace(graphId, traceId).reset_all_symbolPens(update=False)
                howToPlot = self.theDataLink.get_howToPlotGraph(graphId)
                howToPlot.meta = None
                self.theWGPlot.get_trace(graphId, traceId).reset_all_brushes(update=False)
        self.update_graphs()



from .onclickInterface import OnclickInterface
from PyQt5.QtWidgets import QApplication
from optimeed.core import obj_to_json, printIfShown, SHOW_INFO


class Onclick_tojson(OnclickInterface):
    def __init__(self, theDataLink):
        self.theLinkDataGraph = theDataLink

    def graph_clicked(self, theGraphVisual, index_graph, index_trace, indices_points):
        for index_point in indices_points:
            theDevice = self.theLinkDataGraph.get_clicked_item(index_graph, index_trace, index_point)
            theStr = str(obj_to_json(theDevice))
            cb = QApplication.clipboard()
            cb.clear(mode=cb.Clipboard)
            cb.setText(theStr, mode=cb.Clipboard)
            printIfShown("{}\nCopied to clipboard!".format(theStr), SHOW_INFO)

    def get_name(self):
        return "Extract json"



import copy
import os
from abc import abstractmethod
from multiprocessing import Pool, cpu_count

import numpy as np

from optimeed.core import Option_class, Option_int
from optimeed.core.collection import ListDataStruct, AutosaveStruct
from optimeed.core.tools import getPath_workspace, rsetattr, rgetattr


class Parametric_Collection(ListDataStruct):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    # def __str__(self):
    #     return super().__str__() +\
    #            text_format.DARKCYAN + '\nSWEEP PARAMETER:\n' + text_format.END +\
    #            indentParagraph(str(self.analysed_attribute), 1) +\
    #            text_format.DARKCYAN + '\nBASE DATA:\n' + text_format.END + indentParagraph(str(self.reference_data), 1)

    @staticmethod
    def get_extension():
        return '.colp'


class Parametric_parameter:
    """Abstract class for a parametric parameter"""

    def __init__(self, analyzed_attribute, reference_device):
        """
        :param analyzed_attribute: Analyzed attribute of the device
        :param reference_device: reference device
        """
        self.analyzed_attribute = analyzed_attribute
        self.reference_device = reference_device

    @abstractmethod
    def get_values(self):
        pass

    def get_reference_device(self):
        return self.reference_device

    def get_analyzed_attribute(self):
        return self.analyzed_attribute


class Parametric_minmax(Parametric_parameter):
    def __init__(self, analyzed_attribute, reference_device, minValue, maxValue, is_relative=False, npoints=10):
        """
        Parametric between boundaries min and max.
        :param minValue: Minimum value to use
        :param maxValue: Maximum value to use
        :param is_relative: if values are relative: multiplication factor based on current reference attribute value
        :param npoints: number of points to evaluate between range
        """
        super().__init__(analyzed_attribute, reference_device)
        self.is_relative = is_relative
        self.values = np.linspace(minValue, maxValue, npoints)

    def get_values(self):
        values = self.values
        if self.is_relative:
            reference_value = rgetattr(self.get_reference_device(), self.get_analyzed_attribute())
            for i in range(len(values)):
                values[i] *= reference_value
        return self.values


class Parametric_analysis(Option_class):
    NUMBER_OF_CORES = 1

    def __init__(self, theParametricParameter, theCharacterization, filename_collection=None, autosave=False):
        super().__init__()
        if filename_collection is None:
            self.filename_collection = "Param_{}".format(theParametricParameter.get_analyzed_attribute())
        else:
            self.filename_collection = filename_collection

        self.theCharacterization = theCharacterization
        self.theParametricParameter = theParametricParameter

        self.autosave = autosave
        self.my_autosave = None

        self.add_option(self.NUMBER_OF_CORES, Option_int("Number of cores used in evaluation", 1))

    def run(self):
        """Instantiates input arguments for analysis"""

        # Set devices to evaluate
        values = self.theParametricParameter.get_values()
        listOfDevices = list()
        for value in values:
            device = copy.deepcopy(self.theParametricParameter.get_reference_device())
            rsetattr(device, self.theParametricParameter.get_analyzed_attribute(), value)
            listOfDevices.append(device)

        # Collections out:
        theCollection_out = self.initialize_output_collection()

        # Simulates
        pool = Pool(min(cpu_count(), self.get_option_value(self.NUMBER_OF_CORES)))
        for device in pool.imap_unordered(self.evaluate, listOfDevices):
            theCollection_out.add_data(device)

        if self.my_autosave is not None:
            self.my_autosave.stop_autosave()
            self.my_autosave.save()
            self.my_autosave = None

        return theCollection_out

    def evaluate(self, theDevice):
        self.theCharacterization.compute(theDevice)
        return theDevice

    def initialize_output_collection(self):

        theCollection_out = Parametric_Collection()
        if self.autosave:
            filename = os.path.join(getPath_workspace(), self.filename_collection + theCollection_out.get_extension())
            self.my_autosave = AutosaveStruct(theCollection_out, filename)
            self.my_autosave.start_autosave(60)

        return theCollection_out



from optimeed.core import printIfShown, SHOW_WARNING, text_format
from typing import List, Dict


class Base_Option:
    def __init__(self, name, based_value, choices=None):
        self.name = name
        self.value = based_value
        self.choices = choices
        if self.choices is None:
            self.choices = list()

    def get_value(self):
        return self.value

    def get_name(self):
        return self.name

    def set_value(self, value):
        self.value = value

    def get_choices(self):
        return self.choices


class Option_bool(Base_Option):
    name: str
    value: bool

    def set_value(self, value):
        self.value = bool(value)

    def get_choices(self):
        return [True, False]


class Option_str(Base_Option):
    name: str
    value: str

    def set_value(self, value):
        self.value = str(value)


class Option_int(Base_Option):
    name: str
    value: int

    def set_value(self, value):
        self.value = int(value)


class Option_float(Base_Option):
    name: str
    value: float

    def set_value(self, value):
        self.value = float(value)


class Option_dict(Base_Option):
    name: str
    value: dict

    def set_value(self, value):
        self.value = dict(value)


class Option_class:
    options_bool: Dict[int, Option_bool]
    options_str: Dict[int, Option_str]
    options_int: Dict[int, Option_int]
    options_float: Dict[int, Option_float]
    options_dict: Dict[int, Option_dict]

    def __init__(self):
        """
        A high level class that provides means of managing options (and automatically saving them).
        Reinitializing children does not reinitialize options
        Restriction: option_get_value can not be used within a __init__ method.
        """
        if not hasattr(self, "initialized"):
            self.options_bool = dict()
            self.options_str = dict()
            self.options_int = dict()
            self.options_float = dict()
            self.options_dict = dict()
        self.initialized = True

    @staticmethod
    def get_option_attributes():
        return ["options_bool", "options_int", "options_float", "options_str", "options_dict"]

    def add_option(self, idOption, theOption):
        if type(idOption) is not int:
            raise TypeError("Wrong type for id Option")
        if isinstance(theOption, Option_bool):
            self.options_bool[idOption] = theOption
        elif isinstance(theOption, Option_str):
            self.options_str[idOption] = theOption
        elif isinstance(theOption, Option_int):
            self.options_int[idOption] = theOption
        elif isinstance(theOption, Option_float):
            self.options_float[idOption] = theOption
        elif isinstance(theOption, Option_dict):
            self.options_dict[idOption] = theOption
        else:
            printIfShown("Unvalid option type: {}".format(theOption), SHOW_WARNING)

    def get_option_name(self, idOption):
        for option in self._pack_options():
            if idOption in option:
                return option[idOption].get_name()

    def get_option_value(self, idOption):
        for option in self._pack_options():
            if idOption in option:
                return option[idOption].get_value()

    def set_option(self, idOption, value):
        for option in self._pack_options():
            if idOption in option:
                option[idOption].set_value(value)

    def _pack_options(self):
        return [getattr(self, option) for option in self.get_option_attributes()]

    def __str__(self):
        theStr = ''
        for option in self._pack_options():
            if len(option):
                theStr += text_format.BLUE + 'Options'
                for key in option:
                    theStr += '\n◦ {:<15}'.format(self.get_option_name(key)) + '\t-\t' + str(self.get_option_value(key))
                theStr += text_format.END + "\n"
        return theStr





from PyQt5 import QtCore, QtWidgets, QtGui
from .process_mainloop import start_qt_mainloop


class MainWindow(QtWidgets.QMainWindow):
    """
    Main class that spawns a Qt window. Use :meth:`~gui_mainWindow.run` to display it.
    """

    def __init__(self, QtWidgetList, isLight=True, actionOnWindowClosed=None, neverCloseWindow=False, title_window='Awesome Visualisation Tool', size=None):
        """

        :param QtWidgetList: List of QWidgets to display in the window
        :param isLight: Light theme option (bool)
        :param actionOnWindowClosed: Callback before the window closes (method)
        :param neverCloseWindow: Hide the option to kill the window (bool)
        :param title_window: Title of the window (str)
        :param size: Size of the window ([x, y])
        """
        # config
        super().__init__()

        if not isLight:
            p = self.palette()  # To change color
            p.setColor(QtGui.QPalette.Background, (30, 30, 30))
            self.setPalette(p)

        self.mainbox = QtWidgets.QWidget()
        self.setCentralWidget(self.mainbox)
        self.mainbox.setLayout(QtWidgets.QVBoxLayout())
        self.mainbox.layout().setContentsMargins(0, 0, 0, 0)
        self.setWindowTitle(title_window)

        if neverCloseWindow:
            self.setWindowFlag(QtCore.Qt.WindowCloseButtonHint, False)
        self.neverCloseWindow = neverCloseWindow

        # Add widgets - QtPlot
        splitter = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        self.mainbox.layout().addWidget(splitter)
        for widget in QtWidgetList:
            splitter.addWidget(widget)

        if size is not None:
            self.resize(*size)
            # QtCore.QTimer().singleShot(10, lambda: self.resize(*size))

        self.actionOnWindowClosed = actionOnWindowClosed

    def set_actionOnClose(self, actionOnWindowClosed):
        self.actionOnWindowClosed = actionOnWindowClosed

    def closeEvent(self, event):
        if self.actionOnWindowClosed is not None:
            self.actionOnWindowClosed(self, event)
        super().closeEvent(event)

    def run(self, hold=False):
        """Display the window"""
        self.show()
        self.repaint()
        if hold:
            self.hold()

    @staticmethod
    def hold():
        start_qt_mainloop()

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_Escape:
            if not self.neverCloseWindow:
                self.deleteLater()
                self.closeEvent(None)



from optimeed.core import FilledContourPlot, ContourPlot, SurfPlot, MeshPlot, ScatterPlot3, printIfShown, SHOW_WARNING

try:
    from plotly import graph_objects as go
    hasPlotly = True
except ImportError:
    hasPlotly = False
except ImportError:
    printIfShown("3D plots deactivated (impossible to import graph_objects)", SHOW_WARNING)
    hasPlotly = False

if hasPlotly:
    def _do_scatterPlot(theData: ScatterPlot3):
        x, y, z = theData.get_plot_data()

        return go.Scatter3d(x=x, y=y, z=z, mode='markers', name=theData.get_legend(), showlegend=True, marker=dict(
            size=5,
            color=z if theData.get_color() is None else theData.get_color(),
            colorscale='Viridis',  # choose a colorscale
            opacity=1.0
        ))


    def _do_contourPlot(theData: ContourPlot):
        x, y, z = theData.get_plot_data()
        return go.Contour(x=x, y=y, z=z, contours_coloring='lines', line_smoothing=0.85, contours={'showlabels': True})


    def _do_filledcontourPlot(theData: FilledContourPlot):
        x, y, z = theData.get_plot_data()
        return go.Contour(x=x, y=y, z=z, line_smoothing=0.85)


    def _do_surfPlot(theData: SurfPlot):
        x, y, z = theData.get_plot_data()
        return go.Surface(x=x, y=y, z=z)


    def _do_meshPlot(theData: MeshPlot):
        x, y, z = theData.get_plot_data()
        return go.Mesh3d(x=x, y=y, z=z)


    def _get_goPlot(theData):
        if isinstance(theData, SurfPlot):
            return _do_surfPlot(theData)
        elif isinstance(theData, MeshPlot):
            return _do_meshPlot(theData)
        elif isinstance(theData, FilledContourPlot):
            return _do_filledcontourPlot(theData)
        elif isinstance(theData, ContourPlot):
            return _do_contourPlot(theData)
        elif isinstance(theData, ScatterPlot3):
            return _do_scatterPlot(theData)
        else:
            raise NotImplementedError()


    def plot3d(listOfData3D, options_x=None, options_y=None, options_z=None):
        fig = go.Figure(data=list(map(_get_goPlot, listOfData3D)))
        options = {"x": options_x, "y": options_y, "z": options_z}

        for axis in ["x", "y", "z"]:
            if options[axis] is None:
                options[axis] = dict()
            for data3D in listOfData3D:
                if data3D.get_label(axis):
                    options[axis].update({"title": data3D.get_label(axis)})
                if data3D.get_lim(axis) is not None:
                    options[axis].update({"range": data3D.get_lim(axis)})
        fig.update_layout(
            scene=dict(
                xaxis=options["x"],
                yaxis=options["y"],
                zaxis=options["z"], ),
        )
        fig.layout.scene.camera.projection.type = "orthographic"
        fig.show()
        return fig
else:
    def plot3d(*args, **kwargs):
        printIfShown("3D plots deactivated (plotly not installed)", SHOW_WARNING)



from PyQt5 import QtWidgets


class Widget_menuButton(QtWidgets.QMenu):
    """Same as QMenu, but integrates it behind a button more easily."""
    def __init__(self, theParentButton):
        super().__init__()
        self.theButton = theParentButton
        self.theButton.setMenu(self)

    def showEvent(self, QShowEvent):
        p = self.pos()
        geo = self.theButton.geometry()
        self.move(p.x() + geo.width() - self.geometry().width(), p.y())

    def mouseReleaseEvent(self, QMouseEvent):
        action = self.activeAction()
        if action is not None and action.isEnabled():
            action.setEnabled(False)
            super().mouseReleaseEvent(QMouseEvent)
            action.setEnabled(True)
            action.trigger()
        else:
            super().mouseReleaseEvent(QMouseEvent)



from .onselectInterface import OnselectInterface
from optimeed.visualize.graphs.pyqtgraph import mkBrush, mkPen


class Onselect_highlight(OnselectInterface):
    def __init__(self, theLinkDataGraphs, theWgPlot):
        """Highlight point from selection

        :param theLinkDataGraphs: :class:`~optimeed.core.linkDataGraph.LinkDataGraph`
        :param theWgPlot: :class:`~optimeed.visualize.gui.widgets.widget_graphs_visual.widget_graphs_visual`
        """

        self.theLinkDataGraphs = theLinkDataGraphs
        self.theWgPlot = theWgPlot
        self.curr_id = 0
        self.selections = dict()

        # self.previous_selected = list()

    def selector_updated(self, selection_name, the_collection, selected_data, not_selected_data):
        """
        Action to perform once the data have been selected

        :param selection_name: name of the selection (deprecated ?)
        :param the_collection: the collection
        :param selected_data: indices of the data selected
        :param not_selected_data: indices of the data not selected
        :return:
        """
        # Keep track of changes
        id_selection = self.curr_id
        self.selections[id_selection] = list()

        id_collection = self.theLinkDataGraphs.get_idcollection_from_collection(the_collection)
        res, _ = self.theLinkDataGraphs.get_graph_and_trace_from_idCollection(id_collection)
        for idGraph, idTrace in res:
            idPoints_selected = self.theLinkDataGraphs.get_idPoints_from_indices_in_collection(idGraph, idTrace, selected_data)
            idPoints_NOTselected = self.theLinkDataGraphs.get_idPoints_from_indices_in_collection(idGraph, idTrace, not_selected_data)
            traceVisual = self.theWgPlot.get_trace(idGraph, idTrace)
            traceVisual.set_brushes(idPoints_selected, mkBrush(*traceVisual.get_color(), 255), update=False)  # Darkest
            traceVisual.set_symbolPens(idPoints_selected, traceVisual.get_base_symbol_pen(), update=False)  # Clearest
            traceVisual.set_brushes(idPoints_NOTselected, mkBrush(*traceVisual.get_color(), 30), update=False)  # Clearest
            traceVisual.set_symbolPens(idPoints_NOTselected, mkPen(*traceVisual.get_color(), 30), update=False)  # Clearest

            # Keep track of changes
            self.selections[id_selection].append((traceVisual, idPoints_selected, idPoints_NOTselected))

        self.theWgPlot.update_graphs()

        self.curr_id += 1
        return id_selection

    def cancel_selector(self, selection_identifier):
        if selection_identifier in self.selections:
            for traceVisual, idPoints_selected, idPoints_NOTselected in self.selections[selection_identifier]:
                traceVisual.reset_brushes(idPoints_selected, update=False)
                traceVisual.reset_brushes(idPoints_NOTselected, update=False)
                traceVisual.reset_symbolPens(idPoints_selected, update=False)
                traceVisual.reset_symbolPens(idPoints_NOTselected, update=False)
            self.theWgPlot.update_graphs()

    def get_name(self):
        return "Highlight points"



#!/usr/bin/python
"""
Update a simple plot as rapidly as possible to measure speed.
"""

import argparse
from collections import deque
from time import perf_counter

import numpy as np

import pyqtgraph as pg
import pyqtgraph.functions as fn
import pyqtgraph.parametertree as ptree
from pyqtgraph.Qt import QtCore, QtGui, QtWidgets

# defaults here result in the same configuration as the original PlotSpeedTest
parser = argparse.ArgumentParser()
parser.add_argument('--noise', dest='noise', action='store_true')
parser.add_argument('--no-noise', dest='noise', action='store_false')
parser.set_defaults(noise=True)
parser.add_argument('--nsamples', default=5000, type=int)
parser.add_argument('--frames', default=50, type=int)
parser.add_argument('--fsample', default=1000, type=float)
parser.add_argument('--frequency', default=0, type=float)
parser.add_argument('--amplitude', default=5, type=float)
parser.add_argument('--opengl', dest='use_opengl', action='store_true')
parser.add_argument('--no-opengl', dest='use_opengl', action='store_false')
parser.set_defaults(use_opengl=None)
parser.add_argument('--allow-opengl-toggle', action='store_true',
    help="""Allow on-the-fly change of OpenGL setting. This may cause unwanted side effects.
    """)
args = parser.parse_args()

if args.use_opengl is not None:
    pg.setConfigOption('useOpenGL', args.use_opengl)
    pg.setConfigOption('enableExperimental', args.use_opengl)

# don't limit frame rate to vsync
sfmt = QtGui.QSurfaceFormat()
sfmt.setSwapInterval(0)
QtGui.QSurfaceFormat.setDefaultFormat(sfmt)

class MonkeyCurveItem(pg.PlotCurveItem):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.monkey_mode = ''

    def setMethod(self, param, value):
        self.monkey_mode = value

    def paint(self, painter, opt, widget):
        if self.monkey_mode not in ['drawPolyline']:
            return super().paint(painter, opt, widget)

        painter.setRenderHint(painter.RenderHint.Antialiasing, self.opts['antialias'])
        painter.setPen(pg.mkPen(self.opts['pen']))

        if self.monkey_mode == 'drawPolyline':
            painter.drawPolyline(fn.arrayToQPolygonF(self.xData, self.yData))

app = pg.mkQApp("Plot Speed Test")

default_pen = pg.mkPen()

children = [
    dict(name='sigopts', title='Signal Options', type='group', children=[
        dict(name='noise', type='bool', value=args.noise),
        dict(name='nsamples', type='int', limits=[0, None], value=args.nsamples),
        dict(name='frames', type='int', limits=[1, None], value=args.frames),
        dict(name='fsample', title='sample rate', type='float', value=args.fsample, units='Hz'),
        dict(name='frequency', type='float', value=args.frequency, units='Hz'),
        dict(name='amplitude', type='float', value=args.amplitude),
    ]),
    dict(name='useOpenGL', type='bool', value=pg.getConfigOption('useOpenGL'),
        readonly=not args.allow_opengl_toggle),
    dict(name='enableExperimental', type='bool', value=pg.getConfigOption('enableExperimental')),
    dict(name='pen', type='pen', value=default_pen),
    dict(name='antialias', type='bool', value=pg.getConfigOption('antialias')),
    dict(name='connect', type='list', limits=['all', 'pairs', 'finite', 'array'], value='all'),
    dict(name='fill', type='bool', value=False),
    dict(name='skipFiniteCheck', type='bool', value=False),
    dict(name='plotMethod', title='Plot Method', type='list', limits=['pyqtgraph', 'drawPolyline'])
]

params = ptree.Parameter.create(name='Parameters', type='group', children=children)
pt = ptree.ParameterTree(showHeader=False)
pt.setParameters(params)
pw = pg.PlotWidget()
splitter = QtWidgets.QSplitter()
splitter.addWidget(pt)
splitter.addWidget(pw)
splitter.show()

pw.setWindowTitle('pyqtgraph example: PlotSpeedTest')
pw.setLabel('bottom', 'Index', units='B')
curve = MonkeyCurveItem(pen=default_pen, brush='b')
pw.addItem(curve)

rollingAverageSize = 1000
elapsed = deque(maxlen=rollingAverageSize)

def resetTimings(*args):
    elapsed.clear()

def makeData(*args):
    global data, connect_array, ptr
    sigopts = params.child('sigopts')
    nsamples = sigopts['nsamples']
    frames = sigopts['frames']
    Fs = sigopts['fsample']
    A = sigopts['amplitude']
    F = sigopts['frequency']
    ttt = np.arange(frames * nsamples, dtype=np.float64) / Fs
    data = A*np.sin(2*np.pi*F*ttt).reshape((frames, nsamples))
    if sigopts['noise']:
        data += np.random.normal(size=data.shape)
    connect_array = np.ones(data.shape[-1], dtype=bool)
    ptr = 0
    pw.setRange(QtCore.QRectF(0, -10, nsamples, 20))

def onUseOpenGLChanged(param, enable):
    pw.useOpenGL(enable)

def onEnableExperimentalChanged(param, enable):
    pg.setConfigOption('enableExperimental', enable)

def onPenChanged(param, pen):
    curve.setPen(pen)

def onFillChanged(param, enable):
    curve.setFillLevel(0.0 if enable else None)

params.child('sigopts').sigTreeStateChanged.connect(makeData)
params.child('useOpenGL').sigValueChanged.connect(onUseOpenGLChanged)
params.child('enableExperimental').sigValueChanged.connect(onEnableExperimentalChanged)
params.child('pen').sigValueChanged.connect(onPenChanged)
params.child('fill').sigValueChanged.connect(onFillChanged)
params.child('plotMethod').sigValueChanged.connect(curve.setMethod)
params.sigTreeStateChanged.connect(resetTimings)

makeData()

fpsLastUpdate = perf_counter()
def update():
    global curve, data, ptr, elapsed, fpsLastUpdate

    options = ['antialias', 'connect', 'skipFiniteCheck']
    kwds = { k : params[k] for k in options }
    if kwds['connect'] == 'array':
        kwds['connect'] = connect_array

    # Measure
    t_start = perf_counter()
    curve.setData(data[ptr], **kwds)
    app.processEvents(QtCore.QEventLoop.ProcessEventsFlag.AllEvents)
    t_end = perf_counter()
    elapsed.append(t_end - t_start)
    ptr = (ptr + 1) % data.shape[0]

    # update fps at most once every 0.2 secs
    if t_end - fpsLastUpdate > 0.2:
        fpsLastUpdate = t_end
        average = np.mean(elapsed)
        fps = 1 / average
        pw.setTitle('%0.2f fps - %0.1f ms avg' % (fps, average * 1_000))

timer = QtCore.QTimer()
timer.timeout.connect(update)
timer.start(0)

if __name__ == '__main__':
    pg.exec()
